# BSD 3-Clause License
#
# Adapted from ValentyUSB.
#
# Copyright (c) 2020, Great Scott Gadgets <ktemkin@greatscottgadgets.com>
# Copyright (c) 2018, Luke Valenty
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from amaranth          import Elaboratable, Module, Signal, Cat, Const
from amaranth.lib.cdc  import FFSynchronizer
from amaranth.hdl.xfrm import ResetInserter



class TxShifter(Elaboratable):
    """Transmit Shifter

    TxShifter accepts parallel data and shifts it out serially.

    Parameters
    ----------
    Parameters are passed in via the constructor.

    width : int
        Width of the data to be shifted.

    Input Ports
    -----------
    Input ports are passed in via the constructor.

    i_data: Signal(width)
        Data to be transmitted.

    i_enable: Signal(), input
        When asserted, shifting will be allowed; otherwise, the shifter will be stalled.

    Output Ports
    ------------
    Output ports are data members of the module. All outputs are flopped.

    o_data : Signal()
        Serial data output.

    o_empty : Signal()
        Asserted the cycle before the shifter loads in more i_data.

    o_get : Signal()
        Asserted the cycle after the shifter loads in i_data.

    """
    def __init__(self, width):
        self._width = width

        #
        # I/O Port
        #
        self.i_data   = Signal(width)
        self.i_enable = Signal()
        self.i_clear  = Signal()

        self.o_get    = Signal()
        self.o_empty  = Signal()

        self.o_data   = Signal()


    def elaborate(self, platform):
        m = Module()

        shifter = Signal(self._width)
        pos = Signal(self._width, init=0b1)


        with m.If(self.i_enable):
            empty = Signal()
            m.d.usb += [
                pos.eq(pos >> 1),
                shifter.eq(shifter >> 1),
                self.o_get.eq(empty),
            ]

            with m.If(empty):
                m.d.usb += [
                    shifter.eq(self.i_data),
                    pos.eq(1 << (self._width-1)),
                ]


        with m.If(self.i_clear):
            m.d.usb += [
                shifter.eq(0),
                pos.eq(1)
            ]


        m.d.comb += [
            empty.eq(pos[0]),
            self.o_empty.eq(empty),
            self.o_data.eq(shifter[0]),
        ]

        return m



class TxNRZIEncoder(Elaboratable):
    """
    NRZI Encode

    In order to ensure there are enough bit transitions for a receiver to recover
    the clock usb uses NRZI encoding.  This module processes the incoming
    dj, dk, se0, and valid signals and decodes them to data values.  It
    also pipelines the se0 signal and passes it through unmodified.

    https://www.pjrc.com/teensy/beta/usb20.pdf, USB2 Spec, 7.1.8
    https://en.wikipedia.org/wiki/Non-return-to-zero

    Clock Domain
    ------------
    usb_48 : 48MHz

    Input Ports
    -----------
    i_valid : Signal()
        Qualifies oe, data, and se0.

    i_oe : Signal()
        Indicates that the transmit pipeline should be driving USB.

    i_data : Signal()
        Data bit to be transmitted on USB. Qualified by o_valid.

    i_se0 : Signal()
        Overrides value of o_data when asserted and indicates that SE0 state
        should be asserted on USB. Qualified by o_valid.

    Output Ports
    ------------
    o_usbp : Signal()
        Raw value of USB+ line.

    o_usbn : Signal()
        Raw value of USB- line.

    o_oe : Signal()
        When asserted it indicates that the tx pipeline should be driving USB.
    """

    def __init__(self):
        self.i_valid = Signal()
        self.i_oe = Signal()
        self.i_data = Signal()

        # flop all outputs
        self.o_usbp = Signal()
        self.o_usbn = Signal()
        self.o_oe = Signal()


    def elaborate(self, platform):
        m = Module()

        usbp = Signal()
        usbn = Signal()
        oe = Signal()

        # wait for new packet to start
        with m.FSM(domain="usb_io"):
            with m.State("IDLE"):
                m.d.comb += [
                    usbp.eq(1),
                    usbn.eq(0),
                    oe.eq(0),
                ]

                with m.If(self.i_valid & self.i_oe):
                    # first bit of sync always forces a transition, we idle
                    # in J so the first output bit is K.
                    m.next = "DK"


            # the output line is in state J
            with m.State("DJ"):
                m.d.comb += [
                    usbp.eq(1),
                    usbn.eq(0),
                    oe.eq(1),
                ]

                with m.If(self.i_valid):
                    with m.If(~self.i_oe):
                        m.next = "SE0A"
                    with m.Elif(self.i_data):
                        m.next = "DJ"
                    with m.Else():
                        m.next = "DK"


            # the output line is in state K
            with m.State("DK"):
                m.d.comb += [
                    usbp.eq(0),
                    usbn.eq(1),
                    oe.eq(1),
                ]

                with m.If(self.i_valid):
                    with m.If(~self.i_oe):
                        m.next = "SE0A"
                    with m.Elif(self.i_data):
                        m.next = "DK"
                    with m.Else():
                        m.next = "DJ"


            # first bit of the SE0 state
            with m.State("SE0A"):
                m.d.comb += [
                    usbp.eq(0),
                    usbn.eq(0),
                    oe.eq(1),
                ]

                with m.If(self.i_valid):
                    m.next = "SE0B"

            # second bit of the SE0 state
            with m.State("SE0B"):
                m.d.comb += [
                    usbp.eq(0),
                    usbn.eq(0),
                    oe.eq(1),
                ]

                with m.If(self.i_valid):
                    m.next = "EOPJ"


            # drive the bus back to J before relinquishing control
            with m.State("EOPJ"):
                m.d.comb += [
                    usbp.eq(1),
                    usbn.eq(0),
                    oe.eq(1),
                ]

                with m.If(self.i_valid):
                    m.next = "IDLE"


        m.d.usb_io += [
            self.o_oe.eq(oe),
            self.o_usbp.eq(usbp),
            self.o_usbn.eq(usbn),
        ]

        return m


class TxBitstuffer(Elaboratable):
    """
    Bitstuff Insertion

    Long sequences of 1's would cause the receiver to lose it's lock on the
    transmitter's clock.  USB solves this with bitstuffing.  A '0' is stuffed
    after every 6 consecutive 1's.

    The TxBitstuffer is the only component in the transmit pipeline that can
    delay transmission of serial data.  It is therefore responsible for
    generating the bit_strobe signal that keeps the pipe moving forward.

    https://www.pjrc.com/teensy/beta/usb20.pdf, USB2 Spec, 7.1.9
    https://en.wikipedia.org/wiki/Bit_stuffing

    Clock Domain
    ------------
    usb_12 : 48MHz

    Input Ports
    ------------
    i_data : Signal()
        Data bit to be transmitted on USB.

    Output Ports
    ------------
    o_data : Signal()
        Data bit to be transmitted on USB.

    o_stall : Signal()
        Used to apply backpressure on the tx pipeline.
    """
    def __init__(self):
        self.i_data = Signal()

        self.o_stall = Signal()
        self.o_will_stall = Signal()
        self.o_data = Signal()


    def elaborate(self, platform):
        m = Module()
        stuff_bit = Signal()

        with m.FSM(domain="usb"):

            for i in range(5):

                with m.State(f"D{i}"):
                    # Receiving '1' increments the bitstuff counter.
                    with m.If(self.i_data):
                        m.next = f"D{i+1}"

                    # Receiving '0' resets the bitstuff counter.
                    with m.Else():
                        m.next = "D0"


            with m.State("D5"):
                with m.If(self.i_data):

                    # There's a '1', so indicate we might stall on the next loop.
                    m.d.comb += self.o_will_stall.eq(1),
                    m.next = "D6"

                with m.Else():
                    m.next = "D0"


            with m.State("D6"):
                m.d.comb += stuff_bit.eq(1)
                m.next = "D0"


        m.d.comb += [
            self.o_stall.eq(stuff_bit)
        ]

        # flop outputs
        with m.If(stuff_bit):
            m.d.usb += self.o_data.eq(0),
        with m.Else():
            m.d.usb += self.o_data.eq(self.i_data)

        return m


class TxPipeline(Elaboratable):
    def __init__(self):
        self.i_bit_strobe = Signal()

        self.i_data_payload = Signal(8)
        self.o_data_strobe = Signal()

        self.i_oe = Signal()

        self.o_usbp = Signal()
        self.o_usbn = Signal()
        self.o_oe = Signal()

        self.o_pkt_end = Signal()

        self.fit_dat = Signal()
        self.fit_oe  = Signal()


    def elaborate(self, platform):
        m = Module()

        sync_pulse = Signal(8)

        da_reset_shifter = Signal()
        da_reset_bitstuff = Signal() # Need to reset the bit stuffer 1 cycle after the shifter.
        stall = Signal()

        # These signals are set during the sync pulse
        sp_reset_bitstuff = Signal()
        sp_reset_shifter = Signal()
        sp_bit = Signal()
        sp_o_data_strobe = Signal()

        # 12MHz domain
        bitstuff_valid_data = Signal()

        # Keep a Gray counter around to smoothly transition between states
        state_gray = Signal(2)
        state_data = Signal()
        state_sync = Signal()


        #
        # Transmit gearing.
        #
        m.submodules.shifter = shifter = TxShifter(width=8)
        m.d.comb += [
            shifter.i_data    .eq(self.i_data_payload),

            shifter.i_enable  .eq(~stall),
            shifter.i_clear   .eq(da_reset_shifter | sp_reset_shifter)
        ]

        #
        # Bit-stuffing and NRZI.
        #
        bitstuff = ResetInserter(da_reset_bitstuff)(TxBitstuffer())
        m.submodules.bitstuff = bitstuff

        m.submodules.nrzi = nrzi = TxNRZIEncoder()


        #
        # Transmit controller.
        #

        m.d.comb += [
            # Send a data strobe when we're two bits from the end of the sync pulse.
            # This is because the pipeline takes two bit times, and we want to ensure the pipeline
            # has spooled up enough by the time we're there.
            bitstuff.i_data.eq(shifter.o_data),

            stall.eq(bitstuff.o_stall),

            sp_bit.eq(sync_pulse[0]),
            sp_reset_bitstuff.eq(sync_pulse[0]),

            # The shifter has one clock cycle of latency, so reset it
            # one cycle before the end of the sync byte.
            sp_reset_shifter.eq(sync_pulse[1]),

            sp_o_data_strobe.eq(sync_pulse[5]),

            state_data.eq(state_gray[0] & state_gray[1]),
            state_sync.eq(state_gray[0] & ~state_gray[1]),

            self.fit_oe.eq(state_data | state_sync),
            self.fit_dat.eq((state_data & shifter.o_data & ~bitstuff.o_stall) | sp_bit),
            self.o_data_strobe.eq(state_data & shifter.o_get & ~stall & self.i_oe),
        ]

        # If we reset the shifter, then o_empty will go high on the next cycle.
        #

        m.d.usb += [
            # If the shifter runs out of data, percolate the "reset" signal to the
            # shifter, and then down to the bitstuffer.
            # da_reset_shifter.eq(~stall & shifter.o_empty & ~da_stalled_reset),
            # da_stalled_reset.eq(da_reset_shifter),
            # da_reset_bitstuff.eq(~stall & da_reset_shifter),
            bitstuff_valid_data.eq(~stall & shifter.o_get & self.i_oe),
        ]


        with m.FSM(domain="usb"):

            with m.State('IDLE'):
                with m.If(self.i_oe):
                    m.d.usb += [
                        sync_pulse.eq(1 << 7),
                        state_gray.eq(0b01)
                    ]
                    m.next = "SEND_SYNC"
                with m.Else():
                    m.d.usb += state_gray.eq(0b00)


            with m.State('SEND_SYNC'):
                m.d.usb += sync_pulse.eq(sync_pulse >> 1)

                with m.If(sync_pulse[0]):
                    m.d.usb += state_gray.eq(0b11)
                    m.next = "SEND_DATA"
                with m.Else():
                    m.d.usb += state_gray.eq(0b01)


            with m.State('SEND_DATA'):
                with m.If(~self.i_oe & shifter.o_empty & ~bitstuff.o_stall):
                    with m.If(bitstuff.o_will_stall):
                        m.next = 'STUFF_LAST_BIT'
                    with m.Else():
                        m.d.usb += state_gray.eq(0b10)
                        m.next = 'IDLE'

                with m.Else():
                        m.d.usb += state_gray.eq(0b11)

            with m.State('STUFF_LAST_BIT'):
                m.d.usb += state_gray.eq(0b10)
                m.next = 'IDLE'


        # 48MHz domain
        # NRZI encoding
        nrzi_dat = Signal()
        nrzi_oe = Signal()

        # Cross the data from the 12MHz domain to the 48MHz domain
        cdc_dat = FFSynchronizer(self.fit_dat, nrzi_dat, o_domain="usb_io", stages=3)
        cdc_oe  = FFSynchronizer(self.fit_oe, nrzi_oe, o_domain="usb_io", stages=3)
        m.submodules += [cdc_dat, cdc_oe]

        m.d.comb += [
            nrzi.i_valid.eq(self.i_bit_strobe),
            nrzi.i_data.eq(nrzi_dat),
            nrzi.i_oe.eq(nrzi_oe),

            self.o_usbp.eq(nrzi.o_usbp),
            self.o_usbn.eq(nrzi.o_usbn),
            self.o_oe.eq(nrzi.o_oe),

        ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Full-gateware control request handlers. """

from amaranth               import *

from ..usb2.request         import USBRequestHandler


class ControlRequestHandler(USBRequestHandler):
    """ Pure-gateware USB control request handler. """


    def handle_register_write_request(self, m, new_value_signal, write_strobe, stall_condition=0):
        """ Fills in the current state with a request handler meant to set a register.

        Parameters:
            new_value_signal -- The signal to receive the new value to be applied to the relevant register.
            write_strobe     -- The signal which will be pulsed when new_value_signal contains a update.
            stall_condition  -- If provided, if this condition is true, the request will be STALL'd instead
                                of acknowledged.
            """

        # Provide an response to the STATUS stage.
        with m.If(self.interface.status_requested):

            # If our stall condition is met, stall; otherwise, send a ZLP [USB 8.5.3].
            with m.If(stall_condition):
                m.d.comb += self.interface.handshakes_out.stall.eq(1)
            with m.Else():
                m.d.comb += self.send_zlp()

        # Accept the relevant value after the packet is ACK'd...
        with m.If(self.interface.handshakes_in.ack):
            m.d.comb += [
                write_strobe      .eq(1),
                new_value_signal  .eq(self.interface.setup.value)
            ]

            # ... and then return to idle.
            m.next = 'IDLE'


    def handle_simple_data_request(self, m, transmitter, data, length=1):
        """ Fills in a given current state with a request that returns a given piece of data.

        For e.g. GET_CONFIGURATION and GET_STATUS requests.

        Parameters:
            transmitter -- The transmitter module we're working with.
            data        -- The data to be returned.
        """

        # Connect our transmitter up to the output stream...
        m.d.comb += [
            transmitter.stream          .attach(self.interface.tx),
            Cat(transmitter.data[0:length]).eq(data),
            transmitter.max_length      .eq(length)
        ]

        # ... trigger it to respond when data's requested...
        with m.If(self.interface.data_requested):
            m.d.comb += transmitter.start.eq(1)

        # ... and ACK our status stage.
        with m.If(self.interface.status_requested):
            m.d.comb += self.interface.handshakes_out.ack.eq(1)
            m.next = 'IDLE'



#!/usr/bin/python3
# -*- Mode: Python; py-indent-offset: 4 -*-
#
# Copyright (C) 2005,2007  Ray Burr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

# DISCLAIMER: I don't pretend to be a math wizard.  I don't have a
# deep understanding of all of the theory behind CRCs.  Part of the
# reason I originally wrote this is to help understand and verify CRC
# algorithms in practice.  It is likely that some of my terminology is
# inaccurate.

# Requires at least Python 2.4; tested with 2.4 and 2.5.
"""
This module can model common CRC algorithms given the set of defining
parameters.  This is intended to be easy to use for experimentation
rather than optimized for speed.  It is slow even for a native Python
CRC implementation.

Several common CRC algorithms are predefined in this module.

:authors: Ray Burr
:license: MIT License
:contact: http://www.nightmare.com/~ryb/

Examples
========

  >>> '%X' % CRC32.calcString('123456789')
  'CBF43926'

This test function runs all of the defined algorithms on the test
input string '123456789':

  >>> _printResults()
  CRC-5-USB: 19
  CRC-8-SMBUS: F4
  CRC-15: 059E
  CRC-16: BB3D
  CRC-16-USB: B4C8
  CRC-CCITT: 29B1
  CRC-HDLC: 906E
  CRC-24: 21CF02
  CRC-32: CBF43926
  CRC-32C: E3069283
  CRC-64: 46A5A9388A5BEFFE
  CRC-256: 79B96BDC0C519B239BE759EC0688C86FD25A3F4DF1E7F054AD1F923D0739DAC8

Calculating in parts:

  >>> value = CRC32.calcString('1234')
  >>> '%X' % CRC32.calcString('56789', value)
  'CBF43926'

Or, done a different way:

  >>> crc = CrcRegister(CRC32)
  >>> crc.takeString('1234')
  >>> crc.takeString('56789')
  >>> '%X' % crc.getFinalValue()
  'CBF43926'

Inversion of a CRC function:

  >>> CRC_CCITT.reverse().reflect().calcWord(54321, 16, 0)
  1648
  >>> CRC_CCITT.calcWord(_, 16, 0)
  54321

A 15-bit CRC is used in CAN protocols.  The following sample CAN frame
(in binary here) is converted to hexadecimal for the calcWord call.
The bits after the 15-bit CRC are not included in the CRC::

  0 11101000001 0 0 0 0001 00010010 011000010111011 1 1 1 1111111

This sample CAN frame was found in this paper:
<http://www.anthony-marino.com/documents/HDL_implementation_CAN.pdf>

  >>> '%X' % CRC15.calcWord(0x3A08112, 27)
  '30BB'

If the CRC is included, the remainder should always be zero:

  >>> print(CRC15.calcWord(0x1D0408930BB, 42))
  0

A 5-bit CRC is used some kinds of USB packets.  Here is a sample
start-of-frame packet:

  10100101 01100111000 01111

(found at <http://www.nital.com/corporate/usb2snooper.html>)

The first field is the PID (not included in the CRC), the next 11-bit
field is the frame number (0xE6, LSb-first order), and the final five
bits are the CRC (0x1E, LSb-first order).

  >>> '%X' % CRC5_USB.calcWord(0xE6, 11)
  '1E'
"""

# <http://docutils.sourceforge.net/docs/ref/rst/restructuredtext.html>
__docformat__ = "restructuredtext en"

__version__ = "20070611"


class CrcAlgorithm:
    """
    Represents the parameters of a CRC algorithm.
    """

    # FIXME: Instances are supposed to be immutable, but attributes are
    # writable.

    def __init__(self,
                 width,
                 polynomial,
                 name=None,
                 seed=0,
                 lsbFirst=False,
                 lsbFirstData=None,
                 xorMask=0):
        """
        :param width:

          The number of bits in the CRC register, or equivalently, the
          degree of the polynomial.

        :type width:

          an integer

        :param polynomial:

          The generator polynomial as a sequence of exponents

        :type polynomial:

          sequence or integer

        :param name:

          A name identifying algorithm.

        :type name:

          *str*

        :param seed:

          The initial value to load into the register.  (This is the
          value without *xorMask* applied.)

        :type seed:

          an integer

        :param lsbFirst:

          If ``true``, the register shifts toward the
          least-significant bit (sometimes called the *reflected* or
          *reversed* algorithim).  Otherwise, the register shifts
          toward the most-significant bit.

        :type lsbFirst:

          *bool*

        :param lsbFirstData:

          If ``true``, input data is taken least-significant bit
          first.  Otherwise, data is taken most-significant bit first.
          If ``None`` or not given, the value of *lsbFirst* is used.

        :type lsbFirstData:

          *bool*

        :param xorMask:

          An integer mask indicating which bits should be inverted
          when returning the final result.  This is also used for the
          input value if provided.

        :type xorMask:

          an integer
        """

        if width > 0:
            try:
                polyMask = int(polynomial)
            except TypeError:
                # Guess it is already a sequence of exponents.
                polynomial = list(polynomial)
                polynomial.sort()
                polynomial.reverse()
                polynomial = tuple(polynomial)
            else:
                # Convert a mask to a tuple of exponents.
                if lsbFirst:
                    polyMask = reflect(polyMask, width)
                polynomial = (width, )
                for i in range(width - 1, -1, -1):
                    if (polyMask >> i) & 1:
                        polynomial += (i, )

            if polynomial[:1] != (width, ):
                ValueError("mismatch between width and polynomial degree")

        self.width = width
        self.polynomial = polynomial
        self.name = name
        self.seed = seed
        self.lsbFirst = lsbFirst
        self.lsbFirstData = lsbFirstData
        self.xorMask = xorMask

        if not hasattr(width, "__rlshift__"):
            raise ValueError

        # FIXME: Need more checking of parameters.

    def __repr__(self):
        info = ""
        if self.name is not None:
            info = ' "%s"' % str(self.name)
        result = "<%s.%s%s @ %#x>" % (self.__class__.__module__,
                                      self.__class__.__name__, info, id(self))
        return result

    def calcString(self, s, value=None):
        """
        Calculate the CRC of the 8-bit string *s*.
        """
        r = CrcRegister(self, value)
        r.takeString(s)
        return r.getFinalValue()

    def calcWord(self, word, width, value=None):
        """
        Calculate the CRC of the integer *word* as a sequence of
        *width* bits.
        """
        r = CrcRegister(self, value)
        r.takeWord(word, width)
        return r.getFinalValue()

    def reflect(self):
        """
        Return the algorithm with the bit-order reversed.
        """
        ca = CrcAlgorithm(0, 0)
        ca._initFromOther(self)
        ca.lsbFirst = not self.lsbFirst
        if self.lsbFirstData is not None:
            ca.lsbFirstData = not self.lsbFirstData
        if ca.name:
            ca.name += " reflected"
        return ca

    def reverse(self):
        """
        Return the algorithm with the reverse polynomial.
        """
        ca = CrcAlgorithm(0, 0)
        ca._initFromOther(self)
        ca.polynomial = [(self.width - e) for e in self.polynomial]
        ca.polynomial.sort()
        ca.polynomial.reverse()
        ca.polynomial = tuple(ca.polynomial)
        if ca.name:
            ca.name += " reversed"
        return ca

    def _initFromOther(self, other):
        self.width = other.width
        self.polynomial = other.polynomial
        self.name = other.name
        self.seed = other.seed
        self.lsbFirst = other.lsbFirst
        self.lsbFirstData = other.lsbFirstData
        self.xorMask = other.xorMask


class CrcRegister:
    """
    Holds the intermediate state of the CRC algorithm.
    """
    def __init__(self, crcAlgorithm, value=None):
        """
        :param crcAlgorithm:

          The CRC algorithm to use.

        :type crcAlgorithm:

          `CrcAlgorithm`

        :param value:

          The initial register value to use.  The result previous of a
          previous CRC calculation, can be used here to continue
          calculation with more data.  If this parameter is ``None``
          or not given, the register will be initialized with
          algorithm's default seed value.

        :type value:

          an integer
        """

        self.crcAlgorithm = crcAlgorithm
        p = crcAlgorithm

        self.bitMask = (1 << p.width) - 1

        word = 0
        for n in p.polynomial:
            word |= 1 << n
        self.polyMask = word & self.bitMask

        if p.lsbFirst:
            self.polyMask = reflect(self.polyMask, p.width)

        if p.lsbFirst:
            self.inBitMask = 1 << (p.width - 1)
            self.outBitMask = 1
        else:
            self.inBitMask = 1
            self.outBitMask = 1 << (p.width - 1)

        if p.lsbFirstData is not None:
            self.lsbFirstData = p.lsbFirstData
        else:
            self.lsbFirstData = p.lsbFirst

        self.reset()

        if value is not None:
            self.value = value ^ p.xorMask

    def __str__(self):
        return formatBinaryString(self.value, self.crcAlgorithm.width)

    def reset(self):
        """
        Reset the state of the register with the default seed value.
        """
        self.value = int(self.crcAlgorithm.seed)

    def takeBit(self, bit):
        """
        Process a single input bit.
        """
        outBit = ((self.value & self.outBitMask) != 0)
        if self.crcAlgorithm.lsbFirst:
            self.value >>= 1
        else:
            self.value <<= 1
        self.value &= self.bitMask
        if outBit ^ bool(bit):
            self.value ^= self.polyMask

    def takeWord(self, word, width=8):
        """
        Process a binary input word.

        :param word:

          The input word.  Since this can be a Python ``long``, there
          is no coded limit to the number of bits the word can
          represent.

        :type word:

          an integer

        :param width:

          The number of bits *word* represents.

        :type width:

          an integer
        """
        if self.lsbFirstData:
            bitList = list(range(0, width))
        else:
            bitList = list(range(width - 1, -1, -1))
        for n in bitList:
            self.takeBit((word >> n) & 1)

    def takeString(self, s):
        """
        Process a string as input.  It is handled as a sequence of
        8-bit integers.
        """
        for c in s:
            self.takeWord(ord(c))

    def getValue(self):
        """
        Return the current value of the register as an integer.
        """
        return self.value

    def getFinalValue(self):
        """
        Return the current value of the register as an integer with
        *xorMask* applied.  This can be used after all input data is
        processed to obtain the final result.
        """
        p = self.crcAlgorithm
        return self.value ^ p.xorMask


def reflect(value, width):
    return sum(((value >> x) & 1) << (width - 1 - x) for x in range(width))


def formatBinaryString(value, width):
    return "".join("01" [(value >> i) & 1] for i in range(width - 1, -1, -1))


# Some standard algorithms are defined here.  I believe I was able to
# verify the correctness of each of these in some way (against an
# existing implementation or sample data with a known result).

#: Same CRC algorithm as Python's zlib.crc32
CRC32 = CrcAlgorithm(name="CRC-32",
                     width=32,
                     polynomial=(32, 26, 23, 22, 16, 12, 11, 10, 8, 7, 5, 4, 2,
                                 1, 0),
                     seed=0xFFFFFFFF,
                     lsbFirst=True,
                     xorMask=0xFFFFFFFF)

CRC16 = CrcAlgorithm(name="CRC-16",
                     width=16,
                     polynomial=(16, 15, 2, 0),
                     seed=0x0000,
                     lsbFirst=True,
                     xorMask=0x0000)

#: Used in USB data packets.
CRC16_USB = CrcAlgorithm(name="CRC-16-USB",
                         width=16,
                         polynomial=(16, 15, 2, 0),
                         seed=0xFFFF,
                         lsbFirst=True,
                         xorMask=0xFFFF)

CRC_CCITT = CrcAlgorithm(name="CRC-CCITT",
                         width=16,
                         polynomial=(16, 12, 5, 0),
                         seed=0xFFFF,
                         lsbFirst=False,
                         xorMask=0x0000)

#: This is the algorithm used in X.25 and for the HDLC 2-byte FCS.
CRC_HDLC = CrcAlgorithm(name="CRC-HDLC",
                        width=16,
                        polynomial=(16, 12, 5, 0),
                        seed=0xFFFF,
                        lsbFirst=True,
                        xorMask=0xFFFF)

#: Used in ATM HEC and SMBus.
CRC8_SMBUS = CrcAlgorithm(name="CRC-8-SMBUS",
                          width=8,
                          polynomial=(8, 2, 1, 0),
                          seed=0,
                          lsbFirst=False,
                          xorMask=0)

#: Used in RFC-2440 and MIL STD 188-184.
CRC24 = CrcAlgorithm(name="CRC-24",
                     width=24,
                     polynomial=(24, 23, 18, 17, 14, 11, 10, 7, 6, 5, 4, 3, 1,
                                 0),
                     seed=0xB704CE,
                     lsbFirst=False,
                     xorMask=0)

#: Used in Controller Area Network frames.
CRC15 = CrcAlgorithm(name="CRC-15",
                     width=15,
                     polynomial=(15, 14, 10, 8, 7, 4, 3, 0),
                     seed=0,
                     lsbFirst=False,
                     xorMask=0)

#: Used in iSCSI (RFC-3385); usually credited to Guy Castagnoli.
CRC32C = CrcAlgorithm(name="CRC-32C",
                      width=32,
                      polynomial=(32, 28, 27, 26, 25, 23, 22, 20, 19, 18, 14,
                                  13, 11, 10, 9, 8, 6, 0),
                      seed=0xFFFFFFFF,
                      lsbFirst=True,
                      xorMask=0xFFFFFFFF)

#: CRC used in USB Token and Start-Of-Frame packets
CRC5_USB = CrcAlgorithm(name="CRC-5-USB",
                        width=5,
                        polynomial=(5, 2, 0),
                        seed=0x1F,
                        lsbFirst=True,
                        xorMask=0x1F)

#: ISO 3309
CRC64 = CrcAlgorithm(name="CRC-64",
                     width=64,
                     polynomial=(64, 4, 3, 1, 0),
                     seed=0,
                     lsbFirst=True,
                     xorMask=0)

#: This is just to show off the ability to handle a very wide CRC.
# If this is a standard, I don't know where it is from.  I found the
# polynomial on a web page of an apparent Czech "Lady Killer"
# <http://www.volny.cz/lk77/crc256mmx/>.
POLYNOM256 = 0x82E2443E6320383A20B8A2A0A1EA91A3CCA99A30C5205038349C82AAA3A8FD27
CRC256 = CrcAlgorithm(
    name="CRC-256",
    width=256,
    polynomial=POLYNOM256,
    seed=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
    lsbFirst=True,
    xorMask=0)

# For the following I haven't found complete information and/or have
# no way to verify the result.  I started with the list on Wikipedia
# <http://en.wikipedia.org/wiki/Cyclic_redundancy_check>.
#
# CRC4_ITU = CrcAlgorithm(
#    name         = "CRC-4-ITU",
#    width        = 4,
#    polynomial   = (4, 1, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC5_ITU = CrcAlgorithm(
#    name         = "CRC-5-ITU",
#    width        = 5,
#    polynomial   = (5, 4, 2, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC6_ITU = CrcAlgorithm(
#    name         = "CRC-6-ITU",
#    width        = 6,
#    polynomial   = (6, 1, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC7 = CrcAlgorithm(
#    name         = "CRC-7",
#    width        = 7,
#    polynomial   = (7, 3, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC8_CCITT = CrcAlgorithm(
#    name         = "CRC-8-CCITT",
#    width        = 8,
#    polynomial   = (8, 7, 3, 2, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC8_DALLAS = CrcAlgorithm(
#    name         = "CRC-8-Dallas",
#    width        = 8,
#    polynomial   = (8, 5, 4, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC8 = CrcAlgorithm(
#    name         = "CRC-8",
#    width        = 8,
#    polynomial   = (8, 7, 6, 4, 2, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC8_J1850 = CrcAlgorithm(
#    name         = "CRC-8-J1850",
#    width        = 8,
#    polynomial   = (8, 4, 3, 2, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC10 = CrcAlgorithm(
#    name         = "CRC-10",
#    width        = 10,
#    polynomial   = (10, 9, 5, 4, 1, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC12 = CrcAlgorithm(
#    name         = "CRC-12",
#    width        = 12,
#    polynomial   = (12, 11, 3, 2, 1, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)
#
# CRC64_ECMA182 = CrcAlgorithm(
#    name         = "CRC-64-ECMA-182",
#    width        = 64,
#    polynomial   = (64, 62, 57, 55, 54, 53, 52, 47, 46, 45, 40, 39, 38, 37,
#                    35, 33, 32, 31, 29, 27, 24, 23, 22, 21, 19, 17, 13, 12,
#                    10, 9, 7, 4, 1, 0),
#    seed         = ?,
#    lsbFirst     = ?,
#    xorMask      = ?)


def _callCalcString123456789(v):
    return v.calcString('123456789')


def _printResults(fn=_callCalcString123456789):
    import sys
    d = sys.modules[__name__].__dict__
    algorithms = sorted(
        (v for (k, v) in d.items() if isinstance(v, CrcAlgorithm)),
        key=lambda v: (v.width, v.name))
    for a in algorithms:
        format = ("%%0%dX" % ((a.width + 3) // 4))
        print("%s:" % a.name, end=' ')
        print(format % fn(a))


def _test():
    import doctest
    import sys
    return doctest.testmod(sys.modules[__name__])


if __name__ == "__main__":
    _test()




#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from .utils import \
    LunaGatewareTestCase, \
    LunaUSBGatewareTestCase, \
    LunaSSGatewareTestCase, \
    sync_test_case, \
    usb_domain_test_case, \
    fast_domain_test_case, \
    ss_domain_test_case



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth                                       import Module, Signal

from usb_protocol.emitters.descriptors.microsoft10  import MicrosoftOS10DescriptorCollection
from usb_protocol.types                             import USBRequestRecipient, USBRequestType

from ...usb2.request                                import USBRequestHandler
from .ms_descriptor                                 import GetMicrosoftDescriptorHandlerBlock


class MicrosoftOS10RequestHandler(USBRequestHandler):
    """ A platform-specific handler for Microsoft OS 1.0 requests.

    Parameters
    ----------
    descriptors: MicrosoftOS10DescriptorCollection
        A collection of the platform-specific descriptors to respond to Windows with as requested.
    request_code: 
        Request value defined in the device OS string descriptor (0xEE). This is the byte after 'MSFT100'.
        Also called bMS_VendorCode in Microsoft OS 1.0 descriptor specification.
    max_packet_size
        The maximum packet size for the endpoint associated with this handler.
    """
    def __init__(self, descriptors: MicrosoftOS10DescriptorCollection, request_code=0xee, max_packet_size=64):
        self.descriptors      = descriptors
        self._request_code    = request_code
        self._max_packet_size = max_packet_size

        super().__init__()

    def elaborate(self, platform):
        m = Module()

        # Create convenience aliases for our interface components.
        interface           = self.interface
        setup               = interface.setup
        handshake_generator = interface.handshakes_out
        tx                  = interface.tx

        # Handler for GET_DESCRIPTOR_SET requests.
        m.submodules.ms_descriptor_handler = ms_descriptor_handler = \
            GetMicrosoftDescriptorHandlerBlock(self.descriptors)
        m.d.comb += [
            ms_descriptor_handler.index   .eq(setup.index),
            ms_descriptor_handler.length  .eq(setup.length),
        ]

        #
        # Handlers.
        #
        with m.If(
            (setup.type == USBRequestType.VENDOR) &
            (setup.request == self._request_code) & (
                ((setup.recipient == USBRequestRecipient.DEVICE) & (setup.index == 4)) |
                ((setup.recipient == USBRequestRecipient.INTERFACE) & (setup.index == 5)))
        ):
            m.d.comb += interface.claim.eq(1)

            with m.FSM(domain='usb'):

                # IDLE -- not handling any active request
                with m.State('IDLE'):

                    m.d.usb += [
                        # Start at the beginning of our next / fresh GET_DESCRIPTOR_SET request.
                        ms_descriptor_handler.start_position   .eq(0),

                        # Always start our responses with DATA1 pids, per [USB 2.0: 8.5.3].
                        interface.tx_data_pid                  .eq(1)
                    ]

                    # If we've received a new setup packet, handle it.
                    with m.If(setup.received):
                        m.next = 'GET_MS_DESCRIPTOR'


                # GET_MS_DESCRIPTOR -- The host is trying to request a OS Feature descriptor set
                with m.State('GET_MS_DESCRIPTOR'):
                    # Keep track of whether we've sent a packet we're expecting an ACK to.
                    expecting_ack = Signal()

                    m.d.comb += [
                        ms_descriptor_handler.tx    .attach(tx),
                        handshake_generator.stall   .eq(ms_descriptor_handler.stall),
                    ]

                    with m.If(interface.data_requested):
                        m.d.comb += ms_descriptor_handler.start.eq(1)
                        m.d.usb += expecting_ack.eq(1)

                    # Each time we receive an ACK, advance in our descriptor.
                    # This allows us to send descriptors with >64B of content.
                    with m.If(interface.handshakes_in.ack & expecting_ack):

                        next_start_position = ms_descriptor_handler.start_position + self._max_packet_size
                        m.d.usb += [
                            # We've received an ACK; so mark the section we've sent of the descriptor as
                            # received, and move forward...
                            ms_descriptor_handler.start_position    .eq(next_start_position),

                            # ... and toggle our data PID.
                            self.interface.tx_data_pid              .eq(~self.interface.tx_data_pid),

                            # We've got the ACK we expected.
                            expecting_ack                           .eq(0),
                        ]

                    # ... and ACK our status stage.
                    with m.If(interface.status_requested):
                        m.d.comb += handshake_generator.ack.eq(1)
                        m.next = 'IDLE'
                    
                    # If the requested descriptor doesn't exist, the request is terminated by STALLing the data stage.
                    with m.Elif(ms_descriptor_handler.stall):
                        m.d.usb += expecting_ack.eq(0)
                        m.next = 'IDLE'

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Utilities for working with busses. """

import operator
import functools

from amaranth            import Elaboratable, Signal, Module
from amaranth.lib.coding import Encoder


class OneHotMultiplexer(Elaboratable):
    """ Gateware that merges a collection of busses into a single bus.

    The busses joined must meet the following conditions:
        - The relevant type must have a signal that indicates when its data should be
          passed through to the relevant output. This is the 'valid' field.
        - Only one of the relevant busses `valid` fields should be high at a time;
          this effectively makes all of the high signals together a one-hot encoding.
          The implementation's behavior if more than one `valid` signal is undefined.

    I/O port:
        O*: output -- Our output interface; carries the signal merged from all input busses.
    """

    def __init__(self, *, interface_type, valid_field='valid', mux_signals=(), or_signals=(), pass_signals=()):
        """
        Parameters:
            interface_type  -- The type of interface we'll be multiplexing.
            valid_field     -- The name of the field that indicates the relevant object's validity.
            mux_signals     -- An iterable of {signal names to be multiplexed, or functions that
                               accept instances of the relevant interface type and return a Signal}.
                               Signals listed here are passed through iff their one-hot `valid` signal is high.
            or_signals      -- An itereable of {signals names to be multiplexed, or functions that accept
                               an instance of the relevant interface type and return a Signal}. Signals listed
                               here are OR'd together without multiplexing; it's expected that these signals will
                               only be high when their corresponding `valid` signal is high.
            pass_signals    -- A list of signals that should be passed back from the output interface to each
                               of our input interfaces.
        """

        self._valid_field  = valid_field
        self._mux_signals  = mux_signals
        self._or_signals   = or_signals
        self._pass_signals = pass_signals

        # Collection that stores each of the interfaces added to this bus.
        self._inputs = []

        #
        # I/O port
        #
        self.output = interface_type()


    def add_interface(self, input_interface):
        """ Adds an interface to the multiplexer. """
        self._inputs.append(input_interface)


    def add_interfaces(self, interfaces):
        """ Adds a collection/iterable of interfaces to the multiplexer. """
        for interface in interfaces:
            self.add_interface(interface)


    def add_input(self, input_interface):
        """ Alias for add_interface. Adds an interface to the multiplexer. """
        self.add_interface(input_interface)


    @staticmethod
    def _get_signal(interface, name_or_function):
        """ Fetches a signal from the given interface.

        Parameter:
            interface        -- The interface to fetch the relevant signal from.
            name_or_function -- The name of the signal to retrieve; or a function that
                                returns the relevant signal given the interface.
         """

        if callable(name_or_function):
            return name_or_function(interface)
        else:
            return getattr(interface, name_or_function)



    def elaborate(self, platform):
        m = Module()

        #
        # Our module has three core parts:
        #   - an encoder, which converts from our one-hot signal to a mux select line
        #   - a multiplexer, which handles multiplexing e.g. payload signals
        #   - a set of OR'ing logic, which joints together our simple or'd signals

        # Create our encoder...
        m.submodules.encoder = encoder = Encoder(len(self._inputs))
        for index, interface in enumerate(self._inputs):

            # ... and tie its inputs to each of our 'valid' signals.
            valid_signal = getattr(interface, self._valid_field)
            m.d.comb += encoder.i[index].eq(valid_signal)


        # Create our multiplexer, and drive each of our output signals from it.
        with m.Switch(encoder.o):
            for index, interface in enumerate(self._inputs):

                # If an interface is selected...
                with m.Case(index):
                    for identifier in self._mux_signals:

                        # ... connect all of its muxed signals through to the output.
                        output_signal = self._get_signal(self.output, identifier)
                        input_signal  = self._get_signal(interface,   identifier)
                        m.d.comb += output_signal.eq(input_signal)


        # Create the OR'ing logic for each of or or_signals.
        for identifier in self._or_signals:

            # Figure out the signals we want to work with...
            output_signal = self._get_signal(self.output, identifier)
            input_signals = (self._get_signal(i, identifier) for i in self._inputs)

            # ... and OR them together.
            or_reduced = functools.reduce(operator.__or__, input_signals, 0)
            m.d.comb += output_signal.eq(or_reduced)


        # Finally, pass each of our pass-back signals from the output interface
        # back to each of our input interfaces.
        for identifier in self._pass_signals:
            output_signal = self._get_signal(self.output, identifier)

            for interface in self._inputs:
                input_signal = self._get_signal(interface, identifier)
                m.d.comb += input_signal.eq(output_signal)


        return m




#
# This file is part of LUNA.
#
# Copyright (c) 2025 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from luna.gateware.test  import (
    LunaUSBGatewareTestCase,
    usb_domain_test_case,
)

from luna.usb2           import (
    USBIsochronousStreamInEndpoint,
    USBIsochronousStreamOutEndpoint,
)


MAX_PACKET_SIZE = 512

class USBIsochronousStreamInEndpointTest(LunaUSBGatewareTestCase):
    FRAGMENT_UNDER_TEST = USBIsochronousStreamInEndpoint
    FRAGMENT_ARGUMENTS  = {'endpoint_number': 1, 'max_packet_size': MAX_PACKET_SIZE}

    def initialize_signals(self):
        # Configure the endpoint.
        yield self.dut.bytes_in_frame.eq(MAX_PACKET_SIZE)

        # Pretend that our host is always targeting our endpoint.
        yield self.dut.interface.tokenizer.endpoint.eq(self.dut._endpoint_number)
        yield self.dut.interface.tokenizer.is_in.eq(1)

    @usb_domain_test_case
    def test_single_packet_in(self):
        dut = self.dut

        producer = dut.stream
        consumer = dut.interface.tx
        data     = [b % 0xff for b in range(1, MAX_PACKET_SIZE + 1)]
        sent     = []

        # Before we see any data, our streams should all be invalid
        self.assertEqual((yield consumer.first), 0)
        self.assertEqual((yield consumer.last), 0)
        self.assertEqual((yield consumer.payload), 0)
        self.assertEqual((yield consumer.ready), 0)
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield producer.payload), 0)
        self.assertEqual((yield producer.ready), 0)
        self.assertEqual((yield producer.valid), 0)

        # Once we start a new frame ...
        yield dut.interface.tokenizer.new_frame.eq(1)
        yield

        # ... but the host hasn't yet requested data from our endpoint;
        # our stream should still be at rest.
        self.assertEqual((yield consumer.first), 0)
        self.assertEqual((yield dut.data_requested), 0)

        # When the host requests data ...
        yield dut.interface.tokenizer.ready_for_response.eq(1)
        yield

        # ... we go out of State(IDLE) and can check that data_requested is strobed.
        self.assertEqual((yield dut.data_requested), 1)

        # Then one cycle later...
        yield

        # ... we will be in State(SEND_DATA) and our consumer stream becomes valid.
        self.assertEqual((yield consumer.first), 1)
        self.assertEqual((yield consumer.last), 0)
        self.assertEqual((yield consumer.valid), 1)

        # Once the producer has data available ...
        yield producer.valid.eq(1)
        yield producer.payload.eq(data[0])

        # ... but we haven't advanced yet ...
        self.assertEqual((yield producer.ready), 0)
        self.assertEqual((yield consumer.ready), 0)
        self.assertEqual((yield consumer.payload), 0x00)

        # ... until our data is accepted.
        yield consumer.ready.eq(1)
        yield
        sent.append((yield consumer.payload))

        # Now we can chack that the transmitter has the first byte ...
        self.assertEqual((yield producer.ready), 1)
        self.assertEqual((yield consumer.payload), data[0])
        self.assertEqual((yield consumer.first), 1)

        # ... before sending the rest of the packet.
        clocks = 0
        for byte in data[1:]:
            clocks += 1
            yield producer.payload.eq(byte)
            yield
            sent.append((yield consumer.payload))
            self.assertEqual((yield consumer.payload), byte)

        # Finally, we can check that we have received the correct
        # amount of data and that this was the last byte.
        self.assertEqual(sent, data)
        self.assertEqual((yield consumer.last), 1)
        self.assertEqual(clocks, len(data) - 1)


class USBIsochronousStreamOutEndpointTest(LunaUSBGatewareTestCase):
    FRAGMENT_UNDER_TEST = USBIsochronousStreamOutEndpoint
    FRAGMENT_ARGUMENTS  = {'endpoint_number': 1, 'max_packet_size': MAX_PACKET_SIZE}

    def initialize_signals(self):
        # Pretend that our host is always targeting our endpoint.
        yield self.dut.interface.tokenizer.endpoint.eq(self.dut._endpoint_number)
        yield self.dut.interface.tokenizer.is_out.eq(1)


    @usb_domain_test_case
    def test_single_packet_out(self):
        dut = self.dut

        producer = dut.interface.rx
        consumer = dut.stream
        data     = [b % 0xff for b in range(1, MAX_PACKET_SIZE + 1)]
        received = []

        # Before we see any data, our streams should all be invalid.
        self.assertEqual((yield consumer.p.data), 0)
        self.assertEqual((yield consumer.p.first), 0)
        self.assertEqual((yield consumer.p.last), 0)
        self.assertEqual((yield consumer.ready), 0)
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield producer.next), 0)
        self.assertEqual((yield producer.payload), 0)
        self.assertEqual((yield producer.valid), 0)

        # Once the producer sends the first byte ...
        yield producer.valid.eq(1)
        yield producer.next.eq(1)
        yield producer.payload.eq(data[0])
        yield

        # ... and only the first byte ...
        yield producer.next.eq(0)
        yield

        # ... we shouldn't see anything in the consumer stream ...
        self.assertEqual((yield consumer.p.first), 0)
        self.assertEqual((yield consumer.p.last), 0)
        self.assertEqual((yield consumer.p.data), 0)

        # ... but even if we were to mark the consumer's stream as ready ...
        yield consumer.ready.eq(1)
        yield

        # ... the consumer stream will still not be valid because
        # we're using a TransactionalizedFIFO that will only commit
        # once the entire packet has been received.
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield consumer.p.first), 0)
        self.assertEqual((yield consumer.p.last), 0)
        self.assertEqual((yield consumer.p.data), 0)

        # So let's send the rest of the packet ...
        yield producer.next.eq(1)
        clocks = 0
        for byte in data[1:]:
            clocks += 1
            yield producer.payload.eq(byte)
            yield

        # ... which should have taken len(data) - 1 cycles because we already sent the first byte.
        self.assertEqual(clocks, len(data) - 1)
        self.assertEqual((yield producer.payload), data[-1])

        # By now the consumer stream would also have picked up the first byte ...
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield consumer.p.first), 1)
        self.assertEqual((yield consumer.p.data), data[0])

        # ... but the stream still won't advance ...
        yield
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield consumer.p.first), 1)
        self.assertEqual((yield consumer.p.data), data[0])

        # ... until we finally mark the packet as complete and invalidate the producer stream.
        yield dut.interface.rx_complete.eq(1)
        yield producer.valid.eq(0)
        yield producer.next.eq(0)
        yield

        # After three clock cycles delay our stream goes finally valid ...
        yield
        yield
        yield
        self.assertEqual((yield consumer.valid), 1)

        # ... and we can now receive the packet.
        clocks = 0
        while (yield consumer.valid) and (yield consumer.p.last) == 0:
            clocks += 1
            received.append((yield consumer.p.data))
            yield

        self.assertEqual(received, data)
        self.assertEqual((yield consumer.p.last), 1)
        self.assertEqual((yield consumer.p.data), data[-1])
        self.assertEqual(clocks, len(data))

        # Finally, let's invalidate the consumer ...
        yield consumer.ready.eq(0)
        yield

        # ... and everything should be over.
        self.assertEqual((yield producer.valid), 0)
        self.assertEqual((yield producer.next), 0)
        self.assertEqual((yield consumer.ready), 0)
        self.assertEqual((yield consumer.valid), 0)
        self.assertEqual((yield consumer.p.first), 0)
        self.assertEqual((yield consumer.p.last), 0)
        self.assertEqual((yield consumer.p.data), 0)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Endpoint interfaces for working with streams.

The endpoint interfaces in this module provide endpoint interfaces suitable for
connecting streams to USB endpoints.
"""

from amaranth import *
from amaranth.lib.memory import Memory
from usb_protocol.types import USBDirection

from ...stream                import SuperSpeedStreamInterface
from ..protocol.endpoint      import SuperSpeedEndpointInterface


class SuperSpeedStreamInEndpoint(Elaboratable):
    """ Endpoint interface that transmits a simple data stream to a host.

    This interface is suitable for a single bulk or interrupt endpoint.

    This endpoint interface will automatically generate ZLPs when a stream packet would end without
    a short data packet. If the stream's ``last`` signal is tied to zero, then a continuous stream of
    maximum-length-packets will be sent with no inserted ZLPs.

    This implementation is double buffered; and can store a single packet's worth of data while transmitting
    a second packet. Bursting is currently not supported.


    Attributes
    ----------
    stream: SuperSpeedStreamInterface, input stream
        Full-featured stream interface that carries the data we'll transmit to the host.
    interface: SuperSpeedEndpointInterface
        Communications link to our USB device.


    Parameters
    ----------
    endpoint_number: int
        The endpoint number (not address) this endpoint should respond to.
    max_packet_size: int
        The maximum packet size for this endpoint. Should match the wMaxPacketSize provided in the
        USB endpoint descriptor.
    """

    SEQUENCE_NUMBER_BITS = 5


    def __init__(self, *, endpoint_number, max_packet_size=1024):
        self._endpoint_number = endpoint_number
        self._max_packet_size = max_packet_size

        #
        # I/O port
        #
        self.stream    = SuperSpeedStreamInterface()
        self.interface = SuperSpeedEndpointInterface()


    def elaborate(self, platform):
        m = Module()

        interface      = self.interface
        handshakes_in  = interface.handshakes_in
        handshakes_out = interface.handshakes_out

        # Parameters for later use.
        data_width     = len(self.stream.data)
        bytes_per_word = data_width // 8
        buffer_depth   = self._max_packet_size // bytes_per_word

        #
        # Transciever sequencing.
        #

        # Keep track of the sequence number used as we're transmitting.
        sequence_number = Signal(self.SEQUENCE_NUMBER_BITS)

        # Create a signal equal to the next sequence number; for easy comparisons.
        next_sequence_number = Signal.like(sequence_number)
        m.d.comb += next_sequence_number.eq(sequence_number + 1)

        # Advance the sequence number after transmission, or reset it when the endpoint is reset.
        advance_sequence = Signal()
        with m.If(interface.ep_reset):
            m.d.ss += sequence_number.eq(0)
        with m.Elif(advance_sequence):
            m.d.ss += sequence_number.eq(next_sequence_number)


        #
        # Transmit buffer.
        #
        # Our USB connection imposed a few requirements on our stream:
        # 1) we must be able to transmit packets at a full rate; i.e. ```valid``
        #    must be asserted from the start to the end of our transfer; and
        # 2) we must be able to re-transmit data if a given packet is not ACK'd.
        #
        # Accordingly, we'll buffer a full USB packet of data, and then transmit
        # it once either a) our buffer is full, or 2) the transfer ends (last=1).
        #
        # This implementation is double buffered; so a buffer fill can be pipelined
        # with a transmit.
        #
        ping_pong_toggle = Signal()

        # We'll create two buffers; so we can fill one as we empty the other.
        # Since each buffer will be used for every other transaction, we'll use a simple flag to identify
        # which of our "ping-pong" buffers is currently being targeted.
        buffer = Array(Memory(shape=data_width, depth=buffer_depth, init=[]) for _ in range(2))
        buffer_write_ports = Array(buffer[i].write_port(domain="ss") for i in range(2))
        buffer_read_ports  = Array(buffer[i].read_port(domain="ss") for i in range(2))

        m.submodules.transmit_buffer_0, m.submodules.transmit_buffer_1 = buffer

        # Create values equivalent to the buffer numbers for our read and write buffer; which switch
        # whenever we swap our two buffers.
        write_buffer_number =  ping_pong_toggle
        read_buffer_number  = ~ping_pong_toggle

        # Create a shorthand that refers to the buffer to be filled; and the buffer to send from.
        # We'll call these the Read and Write buffers.
        buffer_write = buffer_write_ports[write_buffer_number]
        buffer_read  = buffer_read_ports[read_buffer_number]

        # Buffer state tracking:
        # - Our ``fill_count`` keeps track of how much data is stored in a given buffer.
        # - Our ``stream_ended`` bit keeps track of whether the stream ended while filling up
        #   the given buffer. This indicates that the buffer cannot be filled further; and, when
        #   ``generate_zlps`` is enabled, is used to determine if the given buffer should end in
        #   a short packet; which determines whether ZLPs are emitted.
        buffer_fill_count   = Array(Signal(range(0, self._max_packet_size + 1)) for _ in range(2))
        buffer_stream_ended = Array(Signal(name=f"stream_ended_in_buffer{i}") for i in range(2))

        # Create shortcuts to active fill_count / stream_ended signals for the buffer being written.
        write_fill_count   = buffer_fill_count[write_buffer_number]
        write_stream_ended = buffer_stream_ended[write_buffer_number]

        # Create shortcuts to the fill_count / stream_ended signals for the packet being sent.
        read_fill_count   = buffer_fill_count[read_buffer_number]
        read_stream_ended = buffer_stream_ended[read_buffer_number]

        # Keep track of our current send position; which determines where we are in the packet.
        send_position  = Signal(range(0, self._max_packet_size + 1))

        # Shortcut names.
        in_stream  = self.stream
        out_stream = self.interface.tx

        # We're ready to receive data iff we have space in the buffer we're currently filling.
        m.d.comb += [
            in_stream.ready.eq((write_fill_count + 4 <= self._max_packet_size) & ~write_stream_ended),
            buffer_write.en.eq(in_stream.valid.any() & in_stream.ready),
        ]

        # Increment our fill count whenever we accept new data;
        # based on the number of valid bits we have.
        with m.If(buffer_write.en):
            with m.Switch(in_stream.valid):
                with m.Case(0b0001):
                    m.d.ss += write_fill_count.eq(write_fill_count + 1)
                with m.Case(0b0011):
                    m.d.ss += write_fill_count.eq(write_fill_count + 2)
                with m.Case(0b0111):
                    m.d.ss += write_fill_count.eq(write_fill_count + 3)
                with m.Case(0b1111):
                    m.d.ss += write_fill_count.eq(write_fill_count + 4)

        # If the stream ends while we're adding data to the buffer, mark this as an ended stream.
        with m.If(in_stream.last & buffer_write.en):
            m.d.ss += write_stream_ended.eq(1)


        # Use our memory's two ports to capture data from our transfer stream; and two to emit packets
        # into our packet stream. Since we'll never receive to anywhere else, or transmit to anywhere else,
        # we can just unconditionally connect these.
        m.d.comb += [
            # We'll only ever -write- data from our input stream...
            buffer_write_ports[0].data   .eq(in_stream.payload),
            buffer_write_ports[0].addr   .eq(write_fill_count >> 2),
            buffer_write_ports[1].data   .eq(in_stream.payload),
            buffer_write_ports[1].addr   .eq(write_fill_count >> 2),

            # ... and we'll only ever -send- data from the Read buffer; in the SEND_PACKET state.
            buffer_read.addr             .eq(send_position),
        ]


        #
        # Transmit controller.
        #

        # Stores whether the last packet transmitted was a ZLP. This bit of state determines how
        # retranmission behaves.
        last_packet_was_zlp = Signal()

        # Stores whether we'll need to send an ERDY packet before we send any additional data.
        # If we send an NRDY packet indicating that we have no data for the host, the host will
        # stop polling this endpoint until an ERDY packet is sent [USB3.2r1: 8.10.1]. We'll need
        # to send an ERDY packet to have it resume polling.
        erdy_required = Signal()

        # Shortcut for when we need to deal with an in token.
        # Note that, for USB3, an IN token is an ACK that contains a non-zero ``number_of_packets``.
        is_to_us          = (handshakes_in.endpoint_number == self._endpoint_number)
        is_in_token       = (handshakes_in.number_of_packets != 0)
        ack_received      = handshakes_in.ack_received & is_to_us
        in_token_received = ack_received & is_in_token

        with m.FSM(domain='ss'):

            # WAIT_FOR_DATA -- We don't yet have a full packet to transmit, so  we'll capture data
            # to fill the our buffer. At full throughput, this state will never be reached after
            # the initial post-reset fill.
            with m.State("WAIT_FOR_DATA"):

                # We can't yet send data; so we'll send an NRDY transaction packet.
                with m.If(in_token_received):
                    m.d.comb += handshakes_out.send_nrdy  .eq(1)
                    m.d.ss   += erdy_required             .eq(1)

                # If we have valid data that will end our packet, we're no longer waiting for data.
                # We'll now wait for the host to request data from us.
                packet_complete = (write_fill_count + 4 >= self._max_packet_size)
                will_end_packet = packet_complete | in_stream.last

                with m.If(in_stream.valid & will_end_packet):

                    # If we've just finished a packet, we now have data we can send!
                    with m.If(packet_complete | in_stream.last):
                        m.d.ss += [

                            # We're now ready to take the data we've captured and _transmit_ it.
                            # We'll swap our read and write buffers.
                            ping_pong_toggle.eq(~ping_pong_toggle),

                            # Mark our current stream as no longer having ended.
                            read_stream_ended  .eq(0)
                        ]

                        # If we've already sent an NRDY token, we'll need to request an IN token
                        # before the host will be willing to send us one.
                        with m.If(erdy_required | in_token_received):
                            m.next = "REQUEST_IN_TOKEN"

                        # Otherwise, we can wait for an IN token directly.
                        with m.Else():
                            m.next = "WAIT_TO_SEND"


            # REQUEST_IN_TOKEN -- we now have at least a buffer full of data to send; but
            # we've sent a NRDY token to the host; and thus the host is no longer polling for data.
            # We'll send an ERDY token to the host, in order to request it poll us again.
            with m.State("REQUEST_IN_TOKEN"):

                # Send our ERDY token...
                m.d.comb += handshakes_out.send_erdy.eq(1)

                # ... and once that send is complete, move on to waiting for an IN token.
                with m.If(handshakes_out.done):
                    m.next = "WAIT_TO_SEND"


            # WAIT_TO_SEND -- we now have at least a buffer full of data to send; we'll
            # need to wait for an IN token to send it.
            with m.State("WAIT_TO_SEND"):

                # Once we get an IN token, move to sending a packet.
                with m.If(in_token_received):

                    # If we have a packet to send, send it.
                    with m.If(read_fill_count):
                        m.next = "SEND_PACKET"
                        m.d.ss += [
                            last_packet_was_zlp  .eq(0)
                        ]

                    # Otherwise, we entered a transmit path without any data in the buffer.
                    with m.Else():
                        # ... send a ZLP...
                        m.d.comb += interface.tx_zlp.eq(1)

                        # ... and clear the need to follow up with one, since we've just sent a short packet.
                        m.d.ss += [
                            read_stream_ended    .eq(0),
                            last_packet_was_zlp  .eq(1)
                        ]

                        # We've now completed a packet send; so wait for it to be acknowledged.
                        m.next = "WAIT_FOR_ACK"


            # SEND_PACKET -- we now have enough data to send _and_ have received an IN token.
            # We can now send our data over to the host.
            with m.State("SEND_PACKET"):

                m.d.comb += [
                    # Apply our general transfer information.
                    interface.tx_direction        .eq(USBDirection.IN),
                    interface.tx_sequence_number  .eq(sequence_number),
                    interface.tx_length           .eq(read_fill_count),
                    interface.tx_endpoint_number  .eq(self._endpoint_number),
                ]

                with m.If(~out_stream.valid.any() | out_stream.ready):
                    # Once we emitted a word of data for our receiver, move to the next word in our packet.
                    m.d.ss   += send_position     .eq(send_position + 1)
                    m.d.comb += buffer_read.addr  .eq(send_position + 1)

                    # We're on our last word whenever the next word would be contain the end of our data.
                    first_word = (send_position == 0)
                    last_word  = ((send_position + 1) << 2 >= read_fill_count)

                    m.d.ss += [
                        # Block RAM often has a large clock-to-dout delay; register the output to
                        # improve timings.
                        out_stream.payload        .eq(buffer_read.data),

                        # Let our transmitter know the packet boundaries.
                        out_stream.first          .eq(first_word),
                        out_stream.last           .eq(last_word),
                    ]

                    # Figure out which bytes of our stream are valid. Normally; this is all of them,
                    # but the last word is a special case, which we'll have to handle based on how
                    # many bytes we expect to be valid in the word.
                    with m.If(last_word):

                        # We can figure out how many bytes are valid by looking at the last two bits of our
                        # count; which happen to be the mod-4 remainder.
                        with m.Switch(read_fill_count[0:2]):

                            # If we're evenly divisible by four, all four bytes are valid.
                            with m.Case(0):
                                m.d.ss += out_stream.valid.eq(0b1111)

                            # Otherwise, our remainder tells os how many bytes are valid.
                            with m.Case(1):
                                m.d.ss += out_stream.valid.eq(0b0001)
                            with m.Case(2):
                                m.d.ss += out_stream.valid.eq(0b0011)
                            with m.Case(3):
                                m.d.ss += out_stream.valid.eq(0b0111)


                    # For every word that's not the last one, we know that all bytes are valid.
                    with m.Else():
                        m.d.ss += out_stream.valid.eq(0b1111)

                    # If we've just sent our last word, we're now ready to wait for a response
                    # from our host.
                    with m.If(last_word):
                        m.next = 'WAIT_FOR_ACK'


            # WAIT_FOR_ACK -- We've just sent a packet; but don't know if the host has
            # received it correctly. We'll wait to see if the host ACKs.
            with m.State("WAIT_FOR_ACK"):

                # We're done transmitting data.
                m.d.ss   += out_stream.valid.eq(0)

                # Reset our send-position for the next data packet.
                m.d.ss   += send_position   .eq(0)
                m.d.comb += buffer_read.addr.eq(0)

                # In USB3, an ACK handshake can act as an ACK, an error indicator, and/or an IN token.
                # This helps to maximize bus bandwidth, but means we have to handle each case carefully.
                with m.If(ack_received):

                    # Figure out how the sequence advertisement in our ACK relates to our current sequence number.
                    sequence_advancing = (handshakes_in.next_sequence == next_sequence_number)

                    # Our simplest case is actually when an error occurs, which is indicated by receiving
                    # an ACK packet with Retry set to `1`. For now, we'll also treat a repeated sequence number
                    # as an indication that we need to re-try the given packet.
                    with m.If(handshakes_in.retry_required | ~sequence_advancing):

                        # In this case, we'll re-transmit the relevant data, either by sending another ZLP...
                        with m.If(last_packet_was_zlp):
                            m.d.comb += [
                                interface.tx_zlp.eq(1),
                                advance_sequence.eq(1),
                            ]

                        # ... or by moving right back into sending a data packet.
                        with m.Else():
                            m.next = 'SEND_PACKET'


                    # Otherwise, if our ACK contains the next sequence number, then this is an acknowledgement
                    # of the previous packet [USB3.2r1: 8.12.1.2].
                    with m.Else():

                        # We no longer need to keep the data that's been acknowledged; clear it.
                        m.d.ss += read_fill_count.eq(0)

                        # Figure out if we'll need to follow up with a ZLP. If we have ZLP generation enabled,
                        # we'll make sure we end on a short packet. If this is max-packet-size packet _and_ our
                        # transfer ended with this packet; we'll need to inject a ZLP.
                        follow_up_with_zlp = \
                            (read_fill_count == self._max_packet_size) & read_stream_ended

                        # If we're following up with a ZLP, we have two cases, depending on whether this ACK
                        # is also requesting another packet.
                        with m.If(follow_up_with_zlp):

                            # If we are requesting another packet immediately, we can said ZLP our immediately,
                            # and then continue waiting for the next ACK.
                            with m.If(is_in_token):

                                # ... send a ZLP...
                                m.d.comb += [
                                    interface.tx_zlp.eq(1),
                                    advance_sequence.eq(1),
                                ]

                                # ... and clear the need to follow up with one, since we've just sent a short packet.
                                m.d.ss += [
                                    read_stream_ended    .eq(0),
                                    last_packet_was_zlp  .eq(1)
                                ]

                            # Otherwise, we'll wait for an attempt to send data before we generate a ZLP.
                            with m.Else():
                                m.next = "WAIT_TO_SEND"


                        # Otherwise, there's a possibility we already have a packet-worth of data waiting
                        # for us in our "write buffer", which we've been filling in the background.
                        # If this is the case, we'll flip which buffer we're working with, and then
                        # ready ourselves for transmit.
                        packet_completing = in_stream.valid & (write_fill_count + 4 >= self._max_packet_size)
                        with m.Elif(~in_stream.ready | packet_completing):
                            m.d.comb += [
                                advance_sequence   .eq(1),
                            ]
                            m.d.ss += [
                                ping_pong_toggle   .eq(~ping_pong_toggle),
                                read_stream_ended  .eq(0),
                            ]

                            with m.If(is_in_token):
                                m.d.ss += [
                                    last_packet_was_zlp  .eq(0)
                                ]
                                m.next = "SEND_PACKET"

                            with m.Else():
                                m.next = "WAIT_TO_SEND"

                        # If neither of the above conditions are true; we now don't have enough data to send.
                        # We'll wait for enough data to transmit.
                        with m.Else():
                            m.next = "WAIT_FOR_DATA"

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Control-request interfacing and gateware for USB3. """

from amaranth               import *

from ...request             import SetupPacket
from ...stream              import SuperSpeedStreamInterface
from ..protocol.transaction import HandshakeGeneratorInterface, HandshakeReceiverInterface
from ..protocol.data        import DataHeaderPacket

from ....utils              import falling_edge_detected

class SuperSpeedRequestHandlerInterface:
    """ Interface representing a connection between a control endpoint and a request handler.

    Attributes
    ----------
    setup: SetupPacket(), input to handler
        The setup packet relevant to the current request.
    data_requested: Signal(), input to handler
        Strobe; pulsed when an IN token is received, and thus data is expected.
        If the relevant handler handles this request, it will want to respond with data on ``.tx*``.
    status_requested: Signal(), input to handler
        Strobe; pulsed when a STATUS token is received. The device should respond with a handshake;
        typically either an ACK or a STALL.

    rx: SuperSpeedStreamInterface(), input stream to handler
        A stream containing any data received as part af an OUT request. Its ``ready`` is ignored.
    rx_complete: Signal(), input to handler
        Strobes after successful data receipt. Like most USB data, the packet data cannot be considered
        to be known-good during receipt; but becomes good once :attr:``rx_complete`` is strobed.
    rx_invalid: Signal(), input to handler
        Strobes after an invalid data receipt. Indicates that the most recently received packet was
        corrupted; and should be discarded as invalid.

    tx: SuperSpeedStreamInterface(), output stream from handler
        Should be driven to provide data in response to :attr:``data_requested``. On the same cycle as
        this stream becomes valid, :attr:``tx_length`` and :attr:``tx_sequence_number`` must be valid.
    tx_length: Signal(range(1024 + 1)), output
        The length of the packet to be transmitted on :attr:``tx``. Mandatory for transmission.
    tx_sequence_number: Signal(5)
        The signal number of the packet to be transmitted on :attr:``tx``. Should start with 0, and advance
        by one on each subsequent packet.

    handshakes_out: HandshakeGeneratorInterface(), output from handler
        Carries any handshakes generated by this endpoint. ``endpoint_number`` does not typically need
        to be set; it is driven by the control endpoint.
    handshakes_in: HandshakeGeneratorInterface(), input to handler
        Carries any handshakes received; can be monitored for e.g. ACKs to sent packets.

    address_changed: Signal(), output from handler
        Strobe; should be pulsed when the device's address is to be changed. Should be accompanied by
        a new address on :attr:``new_address``.
    new_address: Signal(7), output from handler
        The new address to be applied to the device when :attr:``address_changed`` is strobed.


    config_changed: Signal(), output from handler
        Strobe; should be pulsed when the device's configuration is to be changed. Should be accompanied by
        a new address on :attr:``new_config``.
    new_config: Signal(8), output from handler
        The new configuration to be applied to the device when :attr:``config_changed`` is strobed.

    current_configuration: Signal(8), input to handler
        The index of the device's current configuration.
    """

    MAX_PACKET_LENGTH = 1024

    def __init__(self):
        # Event signaling.
        self.setup                 = SetupPacket()
        self.data_requested        = Signal()
        self.status_requested      = Signal()

        # Receiver interface.
        self.rx                    = SuperSpeedStreamInterface()
        self.rx_complete           = Signal()
        self.rx_invalid            = Signal()
        self.rx_header             = DataHeaderPacket()

        # Transmitter interface.
        self.tx                    = SuperSpeedStreamInterface()
        self.tx_length             = Signal(range(self.MAX_PACKET_LENGTH + 1))
        self.tx_sequence_number    = Signal(5)

        # Handshake interface.
        self.handshakes_out        = HandshakeGeneratorInterface()
        self.handshakes_in         = HandshakeReceiverInterface()

        # Device state management.
        self.address_changed       = Signal()
        self.new_address           = Signal(7)

        self.active_config         = Signal(8)
        self.config_changed        = Signal()
        self.new_config            = Signal(8)


class SuperSpeedSetupDecoder(Elaboratable):
    """ Gateware that decodes any received Setup packets.

    Attributes
    -----------
    sink: SuperSpeedStreamInterface(), input stream [read-only]
        Packet interface that carres in new data packets. Results should be considered questionable
        until :attr:``packet_good`` or :attr:``packet_bad`` are strobed.

    rx_good: Signal(), input
        Strobe; indicates that the packet received passed validations and can be considered good.
    rx_bad: Signal(), input
        Strobe; indicates that the packet failed CRC checks, or did not end properly.

    header_in: DataHeaderPacket(), input
        Header associated with the active packet.

    packet: SetupPacket(), output
        The parsed contents of our setup packet.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.sink       = SuperSpeedStreamInterface()

        self.rx_good    = Signal()
        self.rx_bad     = Signal()

        self.header_in  = DataHeaderPacket()

        self.packet     = SetupPacket()


    def elaborate(self, platform):
        m = Module()

        # Capture our packet locally, until we have an entire valid packet.
        packet = SetupPacket()

        # Keep our "received" flag low unless explicitly driven.
        m.d.ss += self.packet.received.eq(0)

        with m.FSM(domain="ss"):

            # WAIT_FOR_FIRST -- we're waiting for the first word of a setup packet;
            # which we'll handle on receipt.
            with m.State("WAIT_FOR_FIRST"):
                packet_starting = self.sink.valid.all() & self.sink.first
                packet_is_setup = (self.header_in.setup)

                # Once we see the start of a new setup packet, parse it, and move to the second word.
                with m.If(packet_starting & packet_is_setup):
                    m.d.ss += packet.word_select(0, 32).eq(self.sink.data)
                    m.next = "PARSE_SECOND"

            # PARSE_SECOND -- handle the second and last packet, which contains the remainder of
            # our setup data.
            with m.State("PARSE_SECOND"):

                with m.If(self.sink.valid.all()):

                    # This should be our last word; parse it.
                    with m.If(self.sink.last):
                        m.d.ss += packet.word_select(1, 32).eq(self.sink.data)
                        m.next = "WAIT_FOR_VALID"

                    # If this wasn't our last word, something's gone very wrong.
                    # We'll ignore this packet.
                    with m.Else():
                        m.next = "WAIT_FOR_FIRST"

                # If we see :attr:``rx_bad``, this means our packet aborted early,
                # and thus isn't a valid setup packet. Ignore it, and go back to waiting
                # for our first packet.
                with m.If(self.rx_bad):
                        m.next = "WAIT_FOR_FIRST"

            # WAIT_FOR_VALID -- we've now received all of our data; and we're just waiting
            # for an indication of  whether the data is good or bad.
            with m.State("WAIT_FOR_VALID"):

                # If we see :attr:``packet_good``, this means we have a valid setup packet!
                # We'll output it, and indicate that we've received a new packet.
                with m.If(self.rx_good):
                    m.d.ss += [
                        # Output our stored packet...
                        self.packet           .eq(packet),

                        # ... but strobe its received flag for a cycle.
                        self.packet.received  .eq(1)
                    ]
                    m.next = "WAIT_FOR_FIRST"

                # If we see :attr:``packet_bad``, this means our packet failed CRC checks.
                # We can't do anything with it; so we'll just ignore it.
                with m.If(self.rx_bad):
                    m.next = "WAIT_FOR_FIRST"

        return m


class SuperSpeedRequestHandlerMultiplexer(Elaboratable):
    """ Multiplexes multiple RequestHandlers down to a single interface.

    Interfaces are added using .add_interface().

    Attributes
    ----------
    shared: SuperSpeedRequestHandlerInterface()
        The post-multiplexer RequestHandler interface.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.shared = SuperSpeedRequestHandlerInterface()

        #
        # Internals
        #
        self._interfaces = []


    def add_interface(self, interface: SuperSpeedRequestHandlerInterface):
        """ Adds a RequestHandlerInterface to the multiplexer.

        Arbitration is not performed; it's expected only one handler will be
        driving requests at a time.
        """
        self._interfaces.append(interface)


    def _multiplex_signals(self, m, *, when, multiplex, sub_bus=None):
        """ Helper that creates a simple priority-encoder multiplexer.

        Parmeters:
            when      -- The name of the interface signal that indicates that the `multiplex` signals
                         should be selected for output. If this signals should be multiplex, it
                         should be included in `multiplex`.
            multiplex -- The names of the interface signals to be multiplexed.
        """

        def get_signal(interface, name):
            """ Fetches an interface signal by name / sub_bus. """

            if sub_bus:
                bus = getattr(interface, sub_bus)
                return getattr(bus, name)
            else:
                return  getattr(interface, name)


        # We're building an if-elif tree; so we should start with an If entry.
        conditional = m.If

        for interface in self._interfaces:
            condition = get_signal(interface, when)

            with conditional(condition):

                # Connect up each of our signals.
                for signal_name in multiplex:

                    # Get the actual signals for our input and output...
                    driving_signal = get_signal(interface,   signal_name)
                    target_signal  = get_signal(self.shared, signal_name)

                    # ... and connect them.
                    m.d.comb += target_signal   .eq(driving_signal)

            # After the first element, all other entries should be created with Elif.
            conditional = m.Elif



    def elaborate(self, platform):
        m = Module()
        shared = self.shared

        #
        # Pass through signals being routed -to- our pre-mux interfaces.
        #
        for interface in self._interfaces:
            m.d.comb += [
                # State inputs.
                interface.setup                  .eq(shared.setup),
                interface.active_config          .eq(shared.active_config),

                # Event inputs.
                interface.data_requested         .eq(shared.data_requested),
                interface.status_requested       .eq(shared.status_requested),

                # Receiver inputs.
                interface.rx                     .tap(shared.rx),
                interface.rx_complete            .eq(shared.rx_complete),
                interface.rx_invalid             .eq(shared.rx_invalid),
                interface.rx_header              .eq(shared.rx_header),
                shared.handshakes_in             .connect(interface.handshakes_in),
            ]

        #
        # Multiplex the signals being routed -from- our pre-mux interface.
        #
        self._multiplex_signals(m,
            when='address_changed',
            multiplex=['address_changed', 'new_address']
        )

        self._multiplex_signals(m,
            when='config_changed',
            multiplex=['config_changed', 'new_config']
        )

        #
        # Multiplex each of our transmit interfaces.
        #
        for interface in self._interfaces:

            # If the transmit interface is valid, connect it up to our endpoint.
            # The latest assignment will win; so we can treat these all as a parallel 'if's
            # and still get an appropriate priority encoder.
            with m.If(interface.tx.valid.any()):
                m.d.comb += [
                    shared.tx                  .stream_eq(interface.tx),
                    shared.tx_sequence_number  .eq(interface.tx_sequence_number),
                    shared.tx_length           .eq(interface.tx_length)
                ]


        #
        # Multiplex each of our handshake-out interfaces.
        #
        for interface in self._interfaces:
            any_generate_signal_asserted = (
                interface.handshakes_out.send_ack   |
                interface.handshakes_out.send_stall
            )

            # If the given interface is trying to send an handshake, connect it up
            # to our shared interface.
            with m.If(any_generate_signal_asserted):
                m.d.comb += shared.handshakes_out.connect(interface.handshakes_out)


        return m



class StallOnlyRequestHandler(Elaboratable):
    """ Simple gateware request handler that only conditionally stalls requests.

    I/O port:
        *: interface -- The RequestHandlerInterface used to handle requests.
                        See its record definition for signal definitions.
    """

    def __init__(self, stall_condition):
        """
        Parameters:
            stall_condition -- A function that accepts a SetupRequest packet, and returns
                               an Amaranth conditional indicating whether we should stall.
        """

        self.condition = stall_condition

        #
        # I/O port
        #
        self.interface = SuperSpeedRequestHandlerInterface()


    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        # If we have the opportunity to stall ...
        data_received = falling_edge_detected(m, interface.rx.valid, domain="ss")
        with m.If(interface.data_requested | interface.status_requested | data_received):

            # ... and our stall condition is met ...
            with m.If(self.condition(self.interface.setup)):

                # ... do so.
                m.d.comb += self.interface.handshakes_out.send_stall.eq(1)

        return m



class SuperSpeedRequestHandler(Elaboratable):
    """ Simple base class for request handlers. """

    def __init__(self):
        #
        # I/O port
        #
        self.interface = SuperSpeedRequestHandlerInterface()



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Logical idle detection / polling gateware. """

from amaranth         import *

from ...stream        import USBRawSuperSpeedStream

class IdleHandshakeHandler(Elaboratable):
    """ Unit that performs the USB3 idle handshake.

    The Idle Handshake occurs after link training; and is the first step of
    post-training link initialization.

    Attributes
    ----------
    enable: Signal(), input
        Indicates that the idle handshake has started; which is used to create the
        :attr:``idle_handshake_complete`` signal. It should be asserted once we enter
        Polling.Idle.

    sink: USBRawSuperSpeedStream(), input stream
        Raw post-descrambling stream from the physical layer.

    idle_detected: Signal(), output
        Asserted when the last eight word-aligned symbols detected have been logical
        idle; identifying a completed idle handshake.
    idle_handshake_complete: Signal(), output
        Asserted when we've seen IDLE at least once, and we've been enabled for at least
        16 cycles, as required by the USB3 Idle Handshake [USB3.2r1: 7.5.4.10].
    """

    # We need to send 16B of idle during our handshake. Since we're sending 4B per cycle,
    # that's a total of four cycles.
    RX_CYCLES_REQUIRED = 4

    def __init__(self):

        #
        # I/O port
        #
        self.sink                    = USBRawSuperSpeedStream()

        self.enable                  = Signal()
        self.idle_detected           = Signal()
        self.idle_handshake_complete = Signal()


    def elaborate(self, platform):
        m = Module()

        data_word = self.sink.data
        ctrl_word = self.sink.ctrl

        # Capture the previous data word; so we have a record of eight consecutive signals.
        last_word = Signal.like(data_word)
        last_ctrl = Signal.like(ctrl_word)
        m.d.ss += [
            last_word.eq(data_word),
            last_ctrl.eq(ctrl_word),
        ]

        # Logical idle descrambles to the raw data value zero; so we only need to validate that
        # the last and current words are both zeroes.
        last_word_was_idle   = (last_word == 0) & (last_ctrl == 0)
        current_word_is_idle = (data_word == 0) & (ctrl_word == 0)
        m.d.comb += [
            self.idle_detected  .eq(last_word_was_idle & current_word_is_idle)
        ]

        #
        # Handshake condition detector.
        #
        seen_idle      = Signal()
        enable_counter = Signal(range(self.RX_CYCLES_REQUIRED + 1))

        with m.If(self.enable):

            # Keep track of how many consecutive cycles we're enabled for; as we must
            # send logical idle for at least 16B in order to complete the Idle handshake.
            with m.If(enable_counter < self.RX_CYCLES_REQUIRED):
                m.d.ss += enable_counter.eq(enable_counter + 1)

            # Keep track of whether we've ever seen eight consecutive cycles of idle.
            with m.If(self.idle_detected):
                m.d.ss += seen_idle.eq(1)

            # Our handshake is complete once we've sent logical idle for at least 16 bytes,
            # and we've seen at least eight byte
            send_condition_met = enable_counter == self.RX_CYCLES_REQUIRED
            m.d.comb += self.idle_handshake_complete.eq(seen_idle & send_condition_met)


        # When we're not idle, clear all of our state.
        with m.Else():
            m.d.ss += [
                enable_counter  .eq(0),
                seen_idle       .eq(0)
            ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code adapted from ``litex`` and ``usb3_pipe``.
# SPDX-License-Identifier: BSD-3-Clause
""" Code for USB3 physical-layer encoding. """

from amaranth import *

from .coding   import COM, SHP, SLC, EPF, get_word_for_symbols
from ...stream import USBRawSuperSpeedStream


class RxWordAligner(Elaboratable):
    """ Receiver word alignment.

    Uses the location of COM signals in the data stream to re-position data so that the
    relevant commas always fall in the data's LSB (little endian).
    """

    def __init__(self):

        #
        # I/O port
        #

        # Inputs and outputs.
        self.sink      = USBRawSuperSpeedStream()
        self.source    = USBRawSuperSpeedStream()

        # Debug signals
        self.alignment_offset = Signal(range(4))


    @staticmethod
    def word_meets_alignment_criteria(data, ctrl):
        """ Returns true iff the given data and control appear to be correctly aligned. """

        # Our aligner is simple: we'll search for the start of a TS1/TS2 training set; which we know
        # starts with a burst of four consecutive commas.
        four_comma_data, four_comma_ctrl = get_word_for_symbols(COM, COM, COM, COM)
        return (data == four_comma_data) & (ctrl == four_comma_ctrl)



    def elaborate(self, platform):
        m = Module()

        # Mark ourselves as always ready for new data.
        m.d.comb += self.sink.ready.eq(1)

        # Values from previous cycles.
        previous_data = Signal.like(self.sink.data)
        previous_ctrl = Signal.like(self.sink.ctrl)
        with m.If(self.sink.valid):
            m.d.ss += [
                previous_data  .eq(self.sink.data),
                previous_ctrl  .eq(self.sink.ctrl),
            ]

        # Alignment register: stores how many words the data must be shifted by in order to
        # have correctly aligned data.
        shift_to_apply = Signal(range(4))

        #
        # Alignment shift register.
        #
        # To align our data, we'll create a conglomeration of two consecutive words;
        # and then select the chunk between those words that has the alignment correct.
        data = Cat(previous_data, self.sink.data)
        ctrl = Cat(previous_ctrl, self.sink.ctrl)

        # Create two multiplexers that allow us to select from each of our four possible alignments.
        shifted_data_slices = Array(data[8*i:] for i in range(4))
        shifted_ctrl_slices = Array(ctrl[i:]   for i in range(4))


        #
        # Alignment detection.
        #
        # We'll check each possible alignment to see if it would produce a valid start-of-TS1/TS2;
        # ignoring any words not marked as valid.
        changing_shift = Signal()
        new_shift      = Signal(2)

        with m.If(self.sink.valid):
            possible_alignments = len(shifted_data_slices)
            for i in range(possible_alignments):
                shifted_data = shifted_data_slices[i][0:32]
                shifted_ctrl = shifted_ctrl_slices[i][0: 4]

                # If it would, we'll accept that as our alignment going forward.
                with m.If(self.word_meets_alignment_criteria(shifted_data, shifted_ctrl)):
                    m.d.ss   += shift_to_apply.eq(i)
                    m.d.comb += [
                        changing_shift  .eq(shift_to_apply != i),
                        new_shift       .eq(i)
                    ]


        #
        # Alignment application.
        #

        # Grab the shifted data/ctrl associated with our alignment.

        with m.If(changing_shift):
            m.d.ss += [
                self.source.data       .eq(shifted_data_slices[new_shift]),
                self.source.ctrl       .eq(shifted_ctrl_slices[new_shift]),
                self.source.valid      .eq(self.sink.valid),

                self.alignment_offset  .eq(new_shift),
            ]
        with m.Else():
            m.d.ss += [
                self.source.data       .eq(shifted_data_slices[shift_to_apply]),
                self.source.ctrl       .eq(shifted_ctrl_slices[shift_to_apply]),
                self.source.valid      .eq(self.sink.valid),

                self.alignment_offset  .eq(shift_to_apply),
            ]

        return m



class RxPacketAligner(RxWordAligner):
    """ Receiver word re-alignment.

    Intended to perform a post-descramble (re)-alignment to ensure that we remain aligned to words
    even if the other side does not maintain word alignment. Many links will work without this; but
    the USB3 specification does not guarantee that word-alignment will always be maintained.

    This unit corrects alignment if the link does not maintain this property.
    """


    @staticmethod
    def word_meets_alignment_criteria(data, ctrl):
        """ Returns true iff the given data and control appear to be correctly aligned. """

        # This alignment variant looks for a start of a packet, and attempts to align itself to that.
        # We'll look specifically for the start of Link Commands and Header Packets, as they can start
        # new trains of data. (The start of a Data Packet always follows a header; so anything aligned
        # to a header packet will also be aligned to the data packet.)

        start_of_header,  _  = get_word_for_symbols(SHP, SHP, SHP, EPF)
        start_of_command, _  = get_word_for_symbols(SLC, SLC, SLC, EPF)

        data_matches = (data == start_of_header) | (data == start_of_command)
        ctrl_matches = (ctrl == 0b1111)

        return data_matches & ctrl_matches



#
# This file is part of LUNA.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth                import *
from usb_protocol.emitters   import DeviceDescriptorCollection, SuperSpeedDeviceDescriptorCollection

from luna.usb2               import USBDevice, USBStreamInEndpoint, USBStreamOutEndpoint
from luna.usb3               import USBSuperSpeedDevice, SuperSpeedStreamInEndpoint

from luna.gateware.platform  import NullPin
from luna.gateware.platform.core import LUNAApolloPlatform

from apollo_fpga.gateware.advertiser import ApolloAdvertiser, ApolloAdvertiserRequestHandler

VENDOR_ID  = 0x1209
PRODUCT_ID = 0x0001

BULK_ENDPOINT_NUMBER = 1


class USBSpeedTestDevice(Elaboratable):
    """ Simple device that exchanges data with the host as fast as the hardware can. """

    def __init__(self, generate_clocks=True, fs_only=False, phy_name=None,
            vid=VENDOR_ID, pid=PRODUCT_ID):
        self.generate_clocks = generate_clocks
        self.fs_only = fs_only
        self.phy_name = phy_name
        self.vid = vid
        self.pid = pid
        self.max_bulk_packet_size = 64 if fs_only else 512

    def create_descriptors(self, sharing):
        """ Create the descriptors we want to use for our device. """

        descriptors = DeviceDescriptorCollection()

        #
        # We'll add the major components of the descriptors we we want.
        # The collection we build here will be necessary to create a standard endpoint.
        #

        # We'll need a device descriptor...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = self.vid
            d.idProduct          = self.pid

            d.iManufacturer      = "LUNA"
            d.iProduct           = "speed test"
            d.iSerialNumber      = "no serial"

            d.bNumConfigurations = 1


        # ... and a description of the USB configuration we'll provide.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0

                # Bulk IN to host (tx, from our side)
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self.max_bulk_packet_size

                # Bulk OUT to host (rx, from our side)
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self.max_bulk_packet_size

            if sharing is not None:
                with c.InterfaceDescriptor() as i:
                    i.bInterfaceNumber = 1
                    i.bInterfaceClass = 0xFF
                    i.bInterfaceSubclass = 0x00
                    i.bInterfaceProtocol = ApolloAdvertiserRequestHandler.PROTOCOL_VERSION

        return descriptors


    def elaborate(self, platform):
        m = Module()

        # Generate our domain clocks/resets.
        if self.generate_clocks:
            m.submodules.clocks = platform.clock_domain_generator()

        # Request default PHY unless another was specified.
        phy_name = self.phy_name or platform.default_usb_connection
        ulpi = platform.request(phy_name)

        # Check if port is shared.
        if isinstance(platform, LUNAApolloPlatform):
            sharing = platform.port_sharing(phy_name)
        else:
            sharing = None

        # Create our USB device interface...
        m.submodules.usb = usb = USBDevice(bus=ulpi)

        assert not usb.always_fs or self.fs_only, \
               "fs_only must be set for devices with a full speed only PHY"

        # Add our standard control endpoint to the device.
        descriptors = self.create_descriptors(sharing)
        control_ep = usb.add_standard_control_endpoint(descriptors)

        # Add an Apollo advertiser, if needed.
        if sharing == "advertising":
            adv = m.submodules.adv = ApolloAdvertiser()
            control_ep.add_request_handler(adv.default_request_handler(1))

        # Add a stream endpoint to our device for Bulk IN transfers.
        stream_in_ep = USBStreamInEndpoint(
            endpoint_number=BULK_ENDPOINT_NUMBER,
            max_packet_size=self.max_bulk_packet_size
        )
        usb.add_endpoint(stream_in_ep)

        # Create a simple data source that increments whenever the
        # endpoint is accepting data.
        counter = Signal(8)
        with m.If(stream_in_ep.stream.ready):
            m.d.sync += counter.eq(counter + 1)

        # Send our IN data stream, as fast as we can.
        m.d.comb += [
            stream_in_ep.stream.valid    .eq(1),
            stream_in_ep.stream.payload  .eq(counter)
        ]


        # Add a stream endpoint to our device for Bulk OUT transfers.
        stream_out_ep = USBStreamOutEndpoint(
            endpoint_number=BULK_ENDPOINT_NUMBER,
            max_packet_size=self.max_bulk_packet_size
        )
        usb.add_endpoint(stream_out_ep)

        # Always accept data as it comes in.
        m.d.comb += stream_out_ep.stream.ready.eq(1)

        # Receive our OUT data stream, as fast as we can and output
        # the received data to our User I/O and LEDS
        leds   = Cat(platform.request_optional("led", i, default=NullPin()).o for i in range(6))
        pmod_a = platform.request_optional("user_pmod", 0, default=NullPin(8))
        with m.If(stream_out_ep.stream.valid):
            m.d.comb += [
                leds.eq(stream_out_ep.stream.payload[2:8]),
                pmod_a.o.eq(stream_out_ep.stream.payload),
                pmod_a.oe.eq(1),
            ]


        # Connect our device as a high speed device by default.
        m.d.comb += [
            usb.connect          .eq(1),
            usb.full_speed_only  .eq(1 if self.fs_only else 0),
        ]

        return m


class USBInSuperSpeedTestDevice(Elaboratable):
    """ Simple example of a USB SuperSpeed device using the LUNA framework. """

    MAX_BULK_PACKET_SIZE = 1024

    def __init__(self, generate_clocks=True):
        self.generate_clocks = generate_clocks

    def create_descriptors(self):
        """ Create the descriptors we want to use for our device. """

        descriptors = SuperSpeedDeviceDescriptorCollection()

        #
        # We'll add the major components of the descriptors we we want.
        # The collection we build here will be necessary to create a standard endpoint.
        #

        # We'll need a device descriptor...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = 0x1209
            d.idProduct          = 0x0001

            # We're complying with the USB 3.2 standard.
            d.bcdUSB             = 3.2

            # USB3 requires this to be "9", to indicate 2 ** 9, or 512B.
            d.bMaxPacketSize0    = 9

            d.iManufacturer      = "LUNA"
            d.iProduct           = "SuperSpeed Bulk Test"
            d.iSerialNumber      = "1234"

            d.bNumConfigurations = 1


        # ... and a description of the USB configuration we'll provide.
        with descriptors.ConfigurationDescriptor() as c:
            c.bMaxPower        = 50

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0

                with i.EndpointDescriptor(add_default_superspeed=True) as e:
                    e.bEndpointAddress = 0x80 | BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self.MAX_BULK_PACKET_SIZE

        return descriptors


    def elaborate(self, platform):
        m = Module()

        # Generate our clock domains, if needed.
        if self.generate_clocks:
            m.submodules.clocks = platform.clock_domain_generator()

        # Create our core PIPE PHY. Since PHY configuration is per-board, we'll just ask
        # our platform for a pre-configured USB3 PHY.
        m.submodules.phy = phy = platform.create_usb3_phy()

        # Create our core SuperSpeed device.
        m.submodules.usb = usb = USBSuperSpeedDevice(phy=phy)

        # Add our standard control endpoint to the device.
        descriptors = self.create_descriptors()
        usb.add_standard_control_endpoint(descriptors)

        # Create our example bulk endpoint.
        stream_in_ep = SuperSpeedStreamInEndpoint(
            endpoint_number=BULK_ENDPOINT_NUMBER,
            max_packet_size=self.MAX_BULK_PACKET_SIZE
        )
        usb.add_endpoint(stream_in_ep)

        # Create a simple, monotonically-increasing data stream, and connect that up to
        # to our streaming endpoint.
        counter   = Signal(16)
        stream_in = stream_in_ep.stream

        # Always provide our counter as the input to our stream; it will be consumed
        # whenever our stream endpoint can accept it.
        m.d.comb += [
            stream_in.data    .eq(counter),
            stream_in.valid   .eq(0b1111)
        ]

        # Increment our counter whenever our endpoint is accepting data.
        with m.If(stream_in.ready):
            m.d.ss += counter.eq(counter + 1)

        # Return our elaborated module.
        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Import shortcuts for our most commonly used functionality. """

# Create shorthands for the most common parts of the library's usb2 gateware.
from .gateware.usb.usb2.device                            import USBDevice
from .gateware.usb.usb2.endpoint                          import EndpointInterface
from .gateware.usb.usb2.request                           import RequestHandlerInterface
from .gateware.usb.usb2.endpoints.stream                  import USBStreamInEndpoint, USBStreamOutEndpoint
from .gateware.usb.usb2.endpoints.stream                  import USBMultibyteStreamInEndpoint
from .gateware.usb.usb2.endpoints.status                  import USBSignalInEndpoint
from .gateware.usb.usb2.endpoints.isochronous             import USBIsochronousInEndpoint
from .gateware.usb.usb2.endpoints.isochronous_stream_in   import USBIsochronousStreamInEndpoint
from .gateware.usb.usb2.endpoints.isochronous_stream_out  import USBIsochronousStreamOutEndpoint

__all__ = [
    'USBDevice',
    'EndpointInterface', 'RequestHandlerInterface',
    'USBStreamInEndpoint', 'USBStreamOutEndpoint', 'USBMultibyteStreamInEndpoint',
    'USBSignalInEndpoint', 'USBIsochronousStreamInEndpoint', 'USBIsochronousStreamOutEndpoint',
]



#
# This file is part of LUNA.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Pre-made gateware that implements an ILA connection serial. """

from amaranth                          import Elaboratable, Module, Signal, Cat

from ...debug.ila                      import StreamILA, ILAFrontend
from ...stream                         import StreamInterface
from ..usb2.device                     import USBDevice
from ..usb2.request                    import USBRequestHandler, StallOnlyRequestHandler
from ..usb2.endpoints.stream           import USBMultibyteStreamInEndpoint

from usb_protocol.types                import USBRequestType
from usb_protocol.emitters             import DeviceDescriptorCollection
from usb_protocol.emitters.descriptors import cdc


class USBIntegratedLogicAnalyzer(Elaboratable):
    """ Pre-made gateware that presents a USB-connected ILA.

    Samples are presented over a USB endpoint.
    """

    BULK_ENDPOINT_NUMBER = 1

    def __init__(self, *args, bus=None, delayed_connect=False, max_packet_size=512, **kwargs):
        self._delayed_connect = delayed_connect
        self._max_packet_size = max_packet_size

        # Store our USB bus.
        self._bus = bus

        # Force the ILA's output into the USB domain.
        kwargs['o_domain'] = 'usb'

        # Create our core ILA, which we'll use later.
        self.ila = StreamILA(*args, **kwargs)

        #
        # I/O port
        #

        # Copy some core parameters from our inner ILA.
        self.signals          = self.ila.signals
        self.sample_width     = self.ila.sample_width
        self.sample_depth     = self.ila.sample_depth
        self.sample_rate      = self.ila.sample_rate
        self.sample_period    = self.ila.sample_period
        self.bits_per_sample  = self.ila.bits_per_sample
        self.bytes_per_sample = self.ila.bytes_per_sample

        # Expose our ILA's trigger and status ports directly.
        self.trigger  = self.ila.trigger
        self.sampling = self.ila.sampling
        self.complete = self.ila.complete



    def create_descriptors(self):
        """ Create the descriptors we want to use for our device. """

        descriptors = DeviceDescriptorCollection()

        #
        # We'll add the major components of the descriptors we we want.
        # The collection we build here will be necessary to create a standard endpoint.
        #

        # We'll need a device descriptor...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = 0x1209
            d.idProduct          = 0x0002

            d.iManufacturer      = "LUNA"
            d.iProduct           = "Integrated Logic Analyzer"
            d.iSerialNumber      = "no serial"

            d.bNumConfigurations = 1


        # ... and a description of the USB configuration we'll provide.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | self.BULK_ENDPOINT_NUMBER
                    e.wMaxPacketSize   = self._max_packet_size


        return descriptors


    def elaborate(self, platform):
        m = Module()
        m.submodules.ila = self.ila

        # If we have a bus name rather than a bus object,
        # request the bus from our platform.
        if isinstance(self._bus, str):
            self._bus = platform.request(self._bus)

        # If we have no bus, grab the platform's default USB connection.
        if self._bus is None:
            self._bus = platform.request(platform.default_usb_connection)

        m.submodules.usb = usb = USBDevice(bus=self._bus)

        # Add our standard control endpoint to the device.
        descriptors = self.create_descriptors()
        usb.add_standard_control_endpoint(descriptors)

        # Add a stream endpoint to our device.
        stream_ep = USBMultibyteStreamInEndpoint(
            endpoint_number=self.BULK_ENDPOINT_NUMBER,
            max_packet_size=self._max_packet_size,
            byte_width=self.ila.bytes_per_sample
        )
        usb.add_endpoint(stream_ep)

        # Handle our connection criteria: we'll either connect immediately,
        # or once sampling is done, depending on our _delayed_connect setting.
        connect = Signal()
        if self._delayed_connect:
            with m.If(self.ila.complete):
                m.d.usb += connect.eq(1)
        else:
            m.d.comb += connect.eq(1)

        # Connect up our I/O and our ILA streams.
        m.d.comb += [
            stream_ep.stream  .stream_eq(self.ila.stream),
            usb.connect       .eq(connect)
        ]

        return m



class USBIntegratedLogicAnalyzerFrontend(ILAFrontend):
    """ Frontend for USB-attached integrated logic analyzers.

    Parameters
    ------------
    delay: int
        The number of seconds to wait before trying to connect.
    ila: IntegratedLogicAnalyzer
        The ILA object to work with.
    """

    def __init__(self, *args, ila, delay=3, **kwargs):
        import usb
        import time

        # If we have a connection delay, wait that long.
        if delay:
            time.sleep(delay)

        # Create our USB connection the device
        self._device = usb.core.find(idVendor=0x1209, idProduct=0x0002)


        super().__init__(ila)


    def _split_samples(self, all_samples):
        """ Returns an iterator that iterates over each sample in the raw binary of samples. """
        from apollo_fpga.support.bits import bits

        sample_width_bytes = self.ila.bytes_per_sample

        # Iterate over each sample, and yield its value as a bits object.
        for i in range(0, len(all_samples), sample_width_bytes):
            raw_sample    = all_samples[i:i + sample_width_bytes]
            sample_length = len(Cat(self.ila.signals))

            yield bits.from_bytes(raw_sample, length=sample_length, byteorder='little')


    def _read_samples(self):
        """ Reads a set of ILA samples, and returns them. """

        sample_width_bytes = self.ila.bytes_per_sample
        total_to_read      = self.ila.sample_depth * sample_width_bytes

        # Fetch all of our samples from the given device.
        all_samples = self._device.read(0x81, total_to_read, timeout=0)
        return list(self._split_samples(all_samples))



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
#
# SPDX-License-Identifier: BSD-3-Clause
""" Generating and recognizing LFPS square wave patterns.

SerDes blocks differ in their out-of-band signaling capabilities. Some are capable of detecting
and generating LFPS signaling on their own; others only make it possible to access the high-speed
I/O buffers directly through fabric. This gateware can detect patterns that fit LFPS requriements
given only a bare input buffer, or vice versa.
"""

from math import ceil
from amaranth import *
from amaranth.lib.cdc import FFSynchronizer


__all__ = ['LFPSSquareWaveDetector', 'LFPSSquareWaveGenerator']


# From [USB3.2: Table 6-29]; the maximum and minimum
_LFPS_PERIOD_MIN =  20e-9 # seconds
_LFPS_PERIOD_MAX = 100e-9



class LFPSSquareWaveDetector(Elaboratable):
    """ Detector that identifies LFPS square-wave patterns.

    Operates in the ``pipe`` domain.

    Attributes
    ----------
    rx_gpio: Signal(), input
        The current state of the Rx lines, as retrieved by our SerDes.
    present: Signal(), output
        High whenever we detect an LFPS toggling.
    """

    def __init__(self, pipe_clock_frequency=250e6):

        # Compute the minimum and maximum cycles we're allowed to see.
        # Our multipliers allow for up to a 10% devication in duty cycle.
        self._half_cycle_min = ceil(pipe_clock_frequency * (_LFPS_PERIOD_MIN / 2) * 0.8) - 1
        self._half_cycle_max = ceil(pipe_clock_frequency * (_LFPS_PERIOD_MAX / 2) * 1.2) + 1
        assert self._half_cycle_min >= 1


        #
        # I/O ports
        #
        self.rx_gpio = Signal() # i
        self.present = Signal() # o


    def elaborate(self, platform):
        m = Module()

        # Synchronize the GPIO to our clock domain.
        rx_gpio = Signal()
        m.submodules += FFSynchronizer(self.rx_gpio, rx_gpio, o_domain="pipe")

        # Our mechanism is simple: we measure the length of any periods of consecutive highs and lows
        # we see, and then check to see when they're both in acceptable ranges. Theoretically, we should
        # also check the duty cycle, but as of now, that doesn't seem necessary. [USB3.2: Table 6-29]

        # Keep track of the GPIO's value a cycle ago, so we can easily detect rising and falling edges.
        last_gpio = Signal()
        m.d.pipe += last_gpio.eq(rx_gpio)

        # We'll allow each timer to go one cycle past our half-cycle-max, so it can saturate at an unacceptable
        # level, and mark the ranges as invalid.
        timer_max = self._half_cycle_max + 1

        #
        # Time-high detection.
        #

        # Keep track of our current/total time high.
        current_time_high = Signal(range(0, timer_max + 1))
        total_time_high   = Signal.like(current_time_high)

        # If our GPIO is high, count it.
        with m.If(rx_gpio):

            # Count only when we've reached a value lower than the timer's max,
            # so we saturate once we're outside the acceptable range.
            with m.If(current_time_high != timer_max):
                m.d.pipe += current_time_high.eq(current_time_high + 1)

            # If we've saturated our count, immediately set the total time
            # to the saturation value. This prevents false detections after long
            # strings of constant value.
            with m.Else():
                m.d.pipe += total_time_high.eq(timer_max)


        # If we were still counting last cycle, we'll latch our observed time
        # high before our timer gets cleared. This value represents our total
        # time high, and thus the value we'll use for comparison.
        with m.Elif(last_gpio):
            m.d.pipe += [
                total_time_high    .eq(current_time_high),
                current_time_high  .eq(0)
            ]


        #
        # Time-low detection.
        #

        # Keep track of our current/total time low.
        current_time_low = Signal(range(0, timer_max + 1))
        total_time_low   = Signal.like(current_time_low)

        # If our GPIO is low, count it.
        with m.If(~rx_gpio):

            # Count only when we've reached a value lower than the timer's max,
            # so we saturate once we're outside the acceptable range.
            with m.If(current_time_low != timer_max):
                m.d.pipe += current_time_low.eq(current_time_low + 1)

            # If we've saturated our count, immediately set the total time
            # to the saturation value. This prevents false detections after long
            # strings of constant value.
            with m.Else():
                m.d.pipe += total_time_low.eq(timer_max)

        # If we were still counting last cycle, we'll latch our observed time
        # low before our timer gets cleared. This value represents our total
        # time high, and thus the value we'll use for comparison.
        with m.Elif(~last_gpio):
            m.d.pipe += [
                total_time_low    .eq(current_time_low),
                current_time_low  .eq(0)
            ]


        #
        # Final detection.
        #

        # Whenever both our time high and time low are in range, we have a valid period.
        time_high_valid = ((total_time_high >= self._half_cycle_min) & (total_time_high <= self._half_cycle_max))
        time_low_valid  = ((total_time_low  >= self._half_cycle_min) & (total_time_low  <= self._half_cycle_max))
        m.d.comb += self.present.eq(time_high_valid & time_low_valid)

        return m



class LFPSSquareWaveGenerator(Elaboratable):
    """Generator that outputs LFPS square-wave patterns.
    """
    def __init__(self, lfps_frequency, pipe_clock_frequency):

        # Compute the cycles in one half-period, and make sure the final period is within the spec.
        self._half_cycle = ceil(pipe_clock_frequency / (2 * lfps_frequency))
        assert _LFPS_PERIOD_MIN <= (2 * self._half_cycle) / pipe_clock_frequency <= _LFPS_PERIOD_MAX


        #
        # I/O ports
        #
        self.generate   = Signal() # i

        self.tx_gpio_en = Signal() # o
        self.tx_gpio    = Signal() # o


    def elaborate(self, platform):
        m = Module()

        #
        # LFPS square-wave generator.
        #
        period_timer = Signal(range(self._half_cycle))
        square_wave  = Signal()

        m.d.pipe += period_timer.eq(period_timer + 1)
        with m.If(period_timer + 1 == self._half_cycle):
            m.d.pipe += period_timer.eq(0)
            m.d.pipe += square_wave.eq(~square_wave)

        with m.If(self.generate):
            m.d.comb += [
                self.tx_gpio_en.eq(1),
                self.tx_gpio   .eq(square_wave),
            ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Stream multiplexers/arbiters. """

from amaranth       import *
from .              import StreamInterface


class StreamMultiplexer(Elaboratable):
    """ Gateware that merges a collection of StreamInterfaces into a single interface.

    This variant performs no scheduling. Assumes that only one stream will be communicating at once.

    Attributes
    ----------
    output: StreamInterface(), output stream
        Our output interface; has all of the active busses merged together.
    """

    def __init__(self, stream_type=StreamInterface):
        """
        Parameters:
            stream_type   -- The type of stream we'll be multiplexing. Must be a subclass of StreamInterface.
        """

        # Collection that stores each of the interfaces added to this bus.
        self._inputs = []

        #
        # I/O port
        #
        self.output = stream_type()


    def add_input(self, input_interface):

        """ Adds a transmit interface to the multiplexer. """
        self._inputs.append(input_interface)


    def elaborate(self, platform):
        m = Module()

        #
        # Our basic functionality is simple: we'll build a priority encoder that
        # connects whichever interface has its .valid signal high.
        #

        conditional = m.If

        for interface in self._inputs:

            # If the given interface is asserted, drive our output with its signals.
            with conditional(interface.valid):
                m.d.comb += interface.attach(self.output)

            # After our first iteration, use Elif instead of If.
            conditional = m.Elif


        return m



class StreamArbiter(Elaboratable):
    """ Gateware that merges a collection of StreamInterfaces into a single interface.

    This variant uses a simple priority scheduler; and will use a standard valid/ready handshake
    to schedule a single stream to communicate at a time. Bursts of ``valid`` will never be interrupted,
    so streams will only be switched once the current transmitter drops ``valid`` low.


    Attributes
    ----------
    source: StreamInterface(), output stream
        Our output interface; has all of the active busses merged together.

    idle: Signal(), output
        Asserted when none of our streams is currently active.

    Parameters
    ----------
    stream_type: subclass of StreamInterface
        If provided, sets the type of stream we'll be multiplexing (and thus our output type).
    domain: str
        The name of the domain in which this arbiter should operate. Defaults to "sync".
    """

    def __init__(self, *, stream_type=StreamInterface, domain="sync"):
        self._domain = domain

        # Collection that stores each of the interfaces added to this bus.
        self._sinks = []

        #
        # I/O port
        #
        self.source = stream_type()
        self.idle   = Signal()


    def add_stream(self, stream):
        """ Adds a stream to our arbiter.

        Parameters
        ----------
        stream: StreamInterface subclass
            The stream to be added. Streams added first will have higher priority.
        """
        self._sinks.append(stream)


    def elaborate(self, platform):
        m = Module()
        active_stream = self.source

        # Keep track of which stream is currently active.
        stream_count        = len(self._sinks)
        active_stream_index = Signal(range(stream_count))

        #
        # Stream output multiplexer.
        #
        with m.Switch(active_stream_index):

            # Generate a switch case for each of our possible stream indexes..
            for index, stream in enumerate(self._sinks):
                with m.Case(index):

                    # ... and connect up the stream while in that case.
                    m.d.comb += active_stream.stream_eq(stream)


        #
        # Active stream selection.
        #

        # Only change which stream we're working with when the active stream stops transmitting.
        with m.If(~active_stream.valid):

            # Assume we're idle until proven otherwise.
            m.d.comb += self.idle.eq(1)

            # Check other streams to see if any are valid. We'll use a reversed list in order to maintain
            # our priority order; as the last assignment here "wins".
            for stream_index in reversed(range(stream_count)):

                # If another stream -is- valid, set it to be the active stream.
                with m.If(self._sinks[stream_index].valid):
                    m.d.comb += self.idle.eq(0)
                    m.d.sync += active_stream_index.eq(stream_index)


        # If we're operating in a domain other than sync, replace 'sync' with it.
        if self._domain != "sync":
            m =  DomainRenamer(self._domain)(m)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 physical-layer abstraction."""

import logging

from amaranth import *
from amaranth.lib.fifo import AsyncFIFOBuffered
from ...stream  import USBRawSuperSpeedStream

from .lfps       import LFPSTransceiver
from .scrambling import Scrambler, Descrambler
from .power      import PHYResetController, LinkPartnerDetector
from .ctc        import CTCSkipInserter, CTCSkipRemover
from .alignment  import RxWordAligner, RxPacketAligner

class USB3PhysicalLayer(Elaboratable):
    """ Abstraction encapsulating the USB3 physical layer hardware.

    Performs the lowest-level PHY interfacing, including scrambling/descrambling.

    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input stream
        Data stream accepted from the Link layer; contains raw data to be transmitted.
    source: USBRawSuperSpeedStream(), output stream
        Data stream generated for transit to the Link layer; contains descrambled data accepted from the SerDes.

    enable_scrambling: Signal(), input
        When asserted, scrambling/descrambling will be enabled.
    """

    def __init__(self, *, phy, sync_frequency):
        self._phy = phy
        self._sync_frequency = sync_frequency

        #
        # I/O port
        #

        # Data streams.
        self.sink                       = USBRawSuperSpeedStream()
        self.source                     = USBRawSuperSpeedStream()

        # Raw source (never descrambled; for reset detection).
        self.raw_source                 = USBRawSuperSpeedStream()

        # Physical link state.
        self.ready                      = Signal()
        self.engage_terminations        = Signal()
        self.tx_electrical_idle         = Signal()
        self.invert_rx_polarity         = Signal()
        self.train_equalizer            = Signal()
        self.vbus_present               = Signal()

        # Scrambling control.
        self.enable_scrambling          = Signal()

        # Link partner detection.
        self.perform_rx_detection       = Signal()
        self.link_partner_detected      = Signal()
        self.no_link_partner_detected   = Signal()

        # LFPS control / detection.
        self.send_lfps_polling          = Signal()
        self.lfps_cycles_sent           = Signal(16)

        self.lfps_polling_detected      = Signal()
        self.lfps_reset_detected        = Signal()

        # SKP insertion control.
        self.can_send_skp               = Signal()
        self.skip_removed               = Signal()

        # Debug signaling.
        self.ctc_bytes_in_buffer        = Signal(range(9))
        self.alignment_offset           = Signal(range(4))


    def elaborate(self, platform):
        m = Module()
        phy = self._phy

        #
        # PHY reset & power management.
        #
        m.submodules.reset_controller = reset_controller = PHYResetController(sync_frequency=self._sync_frequency)
        m.d.comb += [
            phy.reset                       .eq(reset_controller.reset),
            reset_controller.phy_status     .eq(phy.phy_status),
            self.ready                      .eq(reset_controller.ready),
        ]


        #
        # PHY control signal handling.
        #
        m.d.comb += [
            # Use USB3.0 5Gbps signaling.
            phy.phy_mode                .eq(0b01),
            phy.rate                    .eq(1),

            # Use nominal half full elastic buffer mode.
            phy.elas_buf_mode           .eq(0),

            # Use default/normal signal thresholds.
            phy.tx_swing                .eq(0),
            phy.tx_margin               .eq(0b000),
            phy.tx_deemph               .eq(0b01),

            # Pass through our remaining control signals directly to the PHY.
            phy.rx_termination          .eq(self.engage_terminations),
            phy.rx_polarity             .eq(self.invert_rx_polarity),
            phy.rx_eq_training          .eq(self.train_equalizer),

            # Pass through our VBUS detected signal up to the link layer.
            self.vbus_present           .eq(phy.power_present)
        ]


        #
        # Link Partner Detection
        #
        m.submodules.rx_detect = rx_detect = LinkPartnerDetector()
        m.d.comb += [
            rx_detect.request_detection    .eq(self.perform_rx_detection),
            rx_detect.phy_status           .eq(phy.phy_status),
            rx_detect.rx_status            .eq(phy.rx_status),

            #self.link_partner_detected     .eq(rx_detect.new_result & rx_detect.partner_present),
            #self.no_link_partner_detected  .eq(rx_detect.new_result & ~rx_detect.partner_present)

            # FIXME: this is temporary; it speeds up partner detection, but isn't strictly correct.
            # (This incorrectly attempts to do link training on USB2 hosts, too).
            phy.power_down                  .eq(0),
            self.link_partner_detected      .eq(phy.power_present)
        ]


        #
        # Transmit output conditioning.
        #

        # Scramble our data before transmitting it, if scrambling is enabled.
        m.submodules.scrambler = scrambler = Scrambler(initial_value=0xffff)
        m.d.comb += [
            scrambler.enable      .eq(self.enable_scrambling),
            scrambler.sink        .stream_eq(self.sink, omit={'valid'}),
            scrambler.sink.valid  .eq(1)
        ]

        # Insert Clock Tolerance Compensation SKP ordered sets, where necessary.
        m.submodules.tx_ctc = tx_ctc = CTCSkipInserter()
        m.d.comb += [
            tx_ctc.sink           .stream_eq(scrambler.source),
            tx_ctc.can_send_skip  .eq(self.can_send_skp),

            scrambler.hold        .eq(tx_ctc.sending_skip)
        ]


        # Convert our Tx stream into PHY connections whenever we're not in electrical idle.
        with m.If(~self.tx_electrical_idle):
            m.d.comb += [
                phy.tx_data          .eq(tx_ctc.source.data),
                phy.tx_datak         .eq(tx_ctc.source.ctrl),
                tx_ctc.source.ready  .eq(1),
            ]


        #
        # Receive input conditioning.
        #

        # Clock tolerance compensation / SKP remover.
        m.submodules.rx_ctc = rx_ctc = CTCSkipRemover()
        m.d.comb += [
            # Connect our PHY's receive data directly to our CTC hardware.
            rx_ctc.sink.data   .eq(phy.rx_data),
            rx_ctc.sink.ctrl   .eq(phy.rx_datak),
            rx_ctc.sink.valid  .eq(1),

            # Diagnostic output.
            self.skip_removed         .eq(rx_ctc.skip_removed),
            self.ctc_bytes_in_buffer  .eq(rx_ctc.bytes_in_buffer),
        ]

        # Word align the data, so it's easily handleable internally.
        m.submodules.aligner = aligner = RxWordAligner()
        m.d.comb += [
            aligner.sink           .stream_eq(rx_ctc.source),
            self.raw_source        .tap(aligner.source),

            self.alignment_offset  .eq(aligner.alignment_offset)
        ]


        # De-scramble our data before output, if needed.
        m.submodules.descrambler = descrambler = Descrambler()
        m.d.comb += [
            descrambler.enable  .eq(self.enable_scrambling),
            descrambler.sink    .stream_eq(aligner.source),
        ]

        # Finally, as a requirement of strict compliance, we'll ensure we can handle packets
        # that arrive even without correct alignment.
        m.submodules.realigner = realigner = RxPacketAligner()
        m.d.comb += [
            realigner.sink      .stream_eq(descrambler.source),
            self.source         .stream_eq(realigner.source),
        ]


        #
        # LFPS signaling.
        #

        m.submodules.lfps_transciever = lfps = LFPSTransceiver()
        m.d.comb += [
            lfps.send_polling           .eq(self.send_lfps_polling),
            self.lfps_cycles_sent       .eq(lfps.cycles_sent),

            self.lfps_polling_detected  .eq(lfps.polling_detected),
            self.lfps_reset_detected    .eq(lfps.reset_detected),

            # The RX_ELECIDLE signal being de-asserted indicates we're receiving valid
            # LFPS signaling. [TUSB1310A: Table 3-3]
            lfps.signaling_received     .eq(~phy.rx_elec_idle),
        ]

        with m.Switch(phy.power_down):

            # In PO, we'll let the LTSSM control electrical idle, and pass through our signals
            # in a way that allows LTSSM.
            with m.Case(0):
                m.d.comb += [
                    # In P0, pass through TX_ELECIDLE directly, as it has its intended meaning.
                    phy.tx_elec_idle              .eq(lfps.drive_electrical_idle | self.tx_electrical_idle),

                    # In P0, the TX_DETRX_LPBK signal is used to drive an LFPS square wave onto the
                    # transmit line when we're in electrical idle, and places us into loopback
                    # otherwise. [TUSB1310A: Table 5-3]  Our LFPS generator takes care of this.
                    phy.tx_detrx_lpbk             .eq(lfps.send_signaling)
                ]

            # For now, we won't support LFPS from states other than P0, as our LTSSM only
            # performs it from P0. We'll use the PHY exclusively for receiver detection.
            with m.Default():
                m.d.comb += [
                    phy.tx_elec_idle              .eq(1),
                    phy.tx_detrx_lpbk             .eq(rx_detect.detection_control)
                ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" Data Packet Payload (DPP) management gateware. """

from amaranth import *

from usb_protocol.types import USBDirection
from usb_protocol.types.superspeed import HeaderPacketType

from .crc              import HeaderPacketCRC, DataPacketPayloadCRC, compute_usb_crc5
from .header           import HeaderPacket, HeaderQueue

from ..physical.coding import SHP, SDP, EPF, stream_matches_symbols
from ...stream         import USBRawSuperSpeedStream, SuperSpeedStreamInterface


class DataHeaderPacket(HeaderPacket):
    DW0_LAYOUT = [
        ('type',                5),
        ('route_string',       20),
        ('device_address',      7),
    ]
    DW1_LAYOUT = [
        ('data_sequence',       5),
        ('reserved_0',          1),
        ('end_of_burst',        1),
        ('direction',           1),
        ('endpoint_number',     4),
        ('reserved_1',          3),
        ('setup',               1),
        ('data_length',        16),
    ]
    DW2_LAYOUT = [
        ('stream_id',          16),
        ('reserved_2',         11),
        ('packet_pending',      1),
        ('reserved_3',          4),
    ]



class DataPacketReceiver(Elaboratable):
    """ Class that monitors the USB bus for data packets, and receives them.

    This class has logic redundant with our Header Packet Receiver, to simplify data packet
    reception. Accordingly, the header section of the data packet will be parsed here as well
    as in the Header Packet receiver. This simplifies our structure at the expense of an additional
    CRC-5 and CRC-16 unit.

    This class performs the validations required at the link layer of the USB specification;
    which include checking the CRC-5 and CRC-16 embedded within the header, and CRC-32 of the
    data packet payload.

    Header sequence number is not checked, here, as a sequence error will force recovery in the
    Header Packet Receiver.


    Attributes
    ----------
    sink: USBRawSuperSpeedStream(), input (monitor only)
        Stream that the USB data to be monitored.

    header: HeaderPacket(), output
        The header packet accompanying the current data packet. Valid once a packet begins
        to be received.
    new_header: Signal(), output
        Strobe; indicates that :attr:``header`` has been updated.
    source: StreamInterface(), output stream
        A stream carrying the data received. Note that the data is not fully validated until
        the packet has been fully received; so this cannot be assumed

    packet_good: Signal(), output
        Strobe; indicates that the packet received passed validations and can be considered good.
    packet_bad: Signal(), output
        Strobe; indicates that the packet failed CRC checks, or did not end properly.
    """

    MAX_PACKET_SIZE = 1024

    def __init__(self):

        #
        # I/O port
        #
        self.sink              = USBRawSuperSpeedStream()

        # Header and data output.
        self.header            = DataHeaderPacket()
        self.new_header        = Signal()
        self.source            = SuperSpeedStreamInterface()

        # State indications.
        self.packet_good       = Signal()
        self.packet_bad        = Signal()



    def elaborate(self, platform):
        m = Module()

        sink   = self.sink
        source = self.source

        # Store our header packet in progress; which we'll output only once it's been validated.
        # We'll store it our generic way; and then refine it as our data becomes valid.
        header = HeaderPacket()

        # Cache our expected CRC5, so we can pipeline generation and comparison.
        expected_crc5 = Signal(5)

        # Store how much data is remaining in the given packet.
        data_bytes_remaining = Signal(range(self.MAX_PACKET_SIZE + 1))

        # Store the most recently received word; which we'll use in case we have an packet which
        # is not evenly divisible into words (e.g. a 3-byte data packet). In these cases, we'll need
        # this previous word for CRC validation, since the final CRC will be partially contained in the
        # last data word.
        previous_word  = Signal.like(self.sink.data)
        previous_valid = Signal.like(self.sink.ctrl)


        #
        # CRC Generators
        #
        m.submodules.crc16 = crc16 = HeaderPacketCRC()
        m.d.comb += crc16.data_input.eq(sink.data),

        m.submodules.crc32 = crc32 = DataPacketPayloadCRC()
        m.d.comb += crc32.data_input.eq(sink.data),

        #
        # Receiver Sequencing
        #
        with m.FSM(domain="ss"):

            # WAIT_FOR_HPSTART -- we're currently waiting for HPSTART framing, which indicates
            # that the following 16 symbols (4 words) will be a header packet.
            with m.State("WAIT_FOR_HPSTART"):

                # Don't start our CRCs until we're past our HPSTART header.
                m.d.comb += [
                    crc16.clear.eq(1),
                    crc32.clear.eq(1),
                ]

                is_hpstart = stream_matches_symbols(sink, SHP, SHP, SHP, EPF)
                with m.If(is_hpstart):
                    m.next = "RECEIVE_DW0"

            # RECEIVE_DWn -- the first three words of our header packet are data words meant form
            # the protocol layer; we'll receive them so we can pass them on to the protocol layer.
            for n in range(3):
                with m.State(f"RECEIVE_DW{n}"):

                    with m.If(sink.valid):
                        m.d.comb += crc16.advance_crc.eq(1)
                        m.d.ss += header[f'dw{n}'].eq(sink.data)
                        m.next = f"RECEIVE_DW{n+1}"

                        # Extra check for our first packet; we'll make sure this of -data- type;
                        # and bail out, otherwise.
                        if n == 0:
                            with m.If(sink.data[0:5] != HeaderPacketType.DATA):
                                m.next = "WAIT_FOR_HPSTART"


            # RECEIVE_DW3 -- we'll receive and parse our final data word, which contains the fields
            # relevant to the link layer.
            with m.State("RECEIVE_DW3"):

                with m.If(sink.valid):
                    m.d.ss += [
                        # Collect the fields from the DW...
                        header.crc16            .eq(sink.data[ 0:16]),
                        header.sequence_number  .eq(sink.data[16:19]),
                        header.dw3_reserved     .eq(sink.data[19:22]),
                        header.hub_depth        .eq(sink.data[22:25]),
                        header.delayed          .eq(sink.data[25]),
                        header.deferred         .eq(sink.data[26]),
                        header.crc5             .eq(sink.data[27:32]),

                        # ... and pipeline a CRC of the to the link control word.
                        expected_crc5           .eq(compute_usb_crc5(sink.data[16:27]))
                    ]

                    m.next = "CHECK_HEADER"

            # CHECK_PACKET -- we've now received our full packet; we'll check it for validity.
            with m.State("CHECK_HEADER"):
                crc5_failed  = (expected_crc5 != header.crc5)
                crc16_failed = (crc16.crc     != header.crc16)

                # If either of our CRCs fail, this isn't going to be followed by a DPP we care about.
                with m.If(crc5_failed | crc16_failed):
                    m.next = "WAIT_FOR_HPSTART"

                # Otherwise, if we have a data packet header, move to capturing our data.
                with m.Elif(stream_matches_symbols(sink, SDP, SDP, SDP, EPF)):
                    m.d.ss += [
                        # Update the header associated with the active packet.
                        self.header           .eq(header),
                        self.new_header       .eq(1),

                        # Read the data length from our header, in preparation to receive it.
                        data_bytes_remaining  .eq(header.dw1[16:]),

                        # Mark the next packet as the first packet in our stream.
                        source.first          .eq(1)
                    ]

                    # Move to receiving data.
                    m.next = "RECEIVE_PAYLOAD"

                # If our data is valid and we're -not- a start of DPP, this isn't for us.
                # Go back to watching for data.
                with m.Elif(sink.valid):
                    m.next = "WAIT_FOR_HPSTART"

            # RECEIVE_PAYLOAD -- receive the core data payload
            with m.State("RECEIVE_PAYLOAD"):
                m.d.comb += [
                    # Pass through most our data directly.
                    source.data         .eq(sink.data),

                    # Manage each of our byte-valid bits directly.
                    source.valid[0]     .eq((data_bytes_remaining > 0) & sink.valid),
                    source.valid[1]     .eq((data_bytes_remaining > 1) & sink.valid),
                    source.valid[2]     .eq((data_bytes_remaining > 2) & sink.valid),
                    source.valid[3]     .eq((data_bytes_remaining > 3) & sink.valid),

                    # Advance our CRC according to how many bytes are currently valid.
                    crc32.advance_word  .eq(source.valid == 0b1111),
                    crc32.advance_3B    .eq(source.valid == 0b0111),
                    crc32.advance_2B    .eq(source.valid == 0b0011),
                    crc32.advance_1B    .eq(source.valid == 0b0001),

                    # Mark this packet as the last one if we've a word or less remaining.
                    source.last         .eq(data_bytes_remaining <= 4)
                ]

                with m.If(sink.valid):
                    # Once we've moved on, this is no longer our first word.
                    m.d.ss += source.first.eq(0)

                    # If we see unexpected control codes in our data packet, bail out.
                    # Note that we'll only check for validity in positions we consider to have
                    # valid data; as we always expect our data packet payload to be followed by
                    # and "end of packet" set of control codes.
                    with m.If((sink.ctrl & source.valid) != 0):
                        m.d.comb += self.packet_bad.eq(1)
                        m.next = "WAIT_FOR_HPSTART"

                    # Capture the current word and valid value, so we can refer to them in
                    # future states. This is necessary for CRC validation when we have a data payload
                    # that's not evenly divisible into words; see the instantiation of ``previous_word``.
                    m.d.ss += [
                        previous_word   .eq(source.data),
                        previous_valid  .eq(source.valid)
                    ]

                    # If we have another word to receive after this, decrement our count,
                    # and continue.
                    with m.If(data_bytes_remaining > 4):
                        m.d.ss += data_bytes_remaining.eq(data_bytes_remaining - 4)

                    with m.Else():
                        m.next = "CHECK_CRC32"


            # CHECK_CRC32 -- we've received the end of our packet; and we're ready to decide if the
            # packet is good or not. We'll check its CRC, and strobe either packet_good or packet_bad.
            with m.State("CHECK_CRC32"):
                data_to_check = Signal.like(sink.data)

                # Depending on how many bytes were present in our data packet, our CRC may be partially
                # contained in the previous word. For example, if we have a 3-byte or 7-byte data packet,
                # one word of the CRC will be contained in the previous data word, and three in our current one.
                with m.Switch(previous_valid):

                    # If our data packet was word aligned, all of our CRC bytes are currently present.
                    # We'll use our current word directly.
                    with m.Case(0b1111):
                        m.d.comb += data_to_check.eq(sink.data)

                    # If we had three valid bytes of data last time, one byte of our CRC was in the previous
                    # word. We'll grab it, and stick it onto the three bytes we're seeing.
                    with m.Case(0b0111):
                        m.d.comb += data_to_check.eq(Cat(previous_word[24:32], sink.data[0:24]))

                    # Same, but for 2B in the previous word and 2B in the current.
                    with m.Case(0b0011):
                        m.d.comb += data_to_check.eq(Cat(previous_word[16:32], sink.data[0:16]))

                    # Same, but for 3B in the previous word and 1B in the current.
                    with m.Case(0b0001):
                        m.d.comb += data_to_check.eq(Cat(previous_word[8:32], sink.data[0:8]))

                # Check our CRC based on the word we've extracted, and strobe either ``packet_good``
                # or ``packet_bad``, depending on its validity.
                with m.If(data_to_check == crc32.crc):
                    m.d.comb += self.packet_good.eq(1)
                with m.Else():
                    m.d.comb += self.packet_bad.eq(1)

                # Finally, wait for our next packet.
                    m.next = "WAIT_FOR_HPSTART"


        return m


class DataPacketTransmitter(Elaboratable):
    """ Gateware that generates a Data Packet Header, and orchestrates sending it and a payload.

    The actual sending is handled by our transmitter gateware.

    Attributes
    ----------
    data_sink: SuperSpeedStreamInterface(), input stream
        The data stream to be send as a data packet. The length of this stream should match thee
        length parameter.
    send_zlp: Signal(), input
        Strobe; triggers sending of a zero-length packet.

    sequence_number: Signal(5), input
        The sequence number associated with the relevant data packet. Latched in once :attr:``data_sink`` goes valid.
    endpoint_number: Signal(4), input
        The endpoint number associated with the relevant data stream. Latched in once :attr:``data_sink`` goes valid.
    data_length: Signal(range(1024 + 1))
        The length of the data packet to be sent; in bytes. Latched in once :attr:``data_sink`` goes valid.
    direction: Signal(), input
        The direction to indicate in the data header packet. Typically Direction.IN; but will be Direction.OUT
        when data is sent to the host as part of a control transfer.

    address: Signal(7), input
        The current address of the USB device.
    """

    MAX_PACKET_SIZE = 1024

    def __init__(self):

        #
        # I/O port
        #

        # Input stream.
        self.data_sink       = SuperSpeedStreamInterface()
        self.send_zlp        = Signal()

        # Data parameters.
        self.sequence_number = Signal(5)
        self.endpoint_number = Signal(4)
        self.data_length     = Signal(range(self.MAX_PACKET_SIZE + 1))
        self.address         = Signal(7)
        self.direction       = Signal()

        # Output streams.
        self.header_source   = HeaderQueue()
        self.data_source     = SuperSpeedStreamInterface()


    def elaborate(self, platform):
        m = Module()

        # Shortcuts.
        header_source = self.header_source
        data_sink     = self.data_sink
        data_source   = self.data_source

        # Latched resources.
        sequence_number = Signal.like(self.sequence_number)
        endpoint_number = Signal.like(self.endpoint_number)
        data_length     = Signal.like(self.data_length)
        direction       = Signal.like(self.direction)


        # For now, we'll pass our data stream through unmodified; only buffered to improve
        # timing.
        #
        # We'll keep this architecture; as later code is likely to want to more actively
        # control when data is passed through to the transmitter.
        with m.If(~data_source.valid.any() | data_source.ready):
            m.d.ss   += data_source.stream_eq(data_sink, omit={'ready'})
            m.d.comb += data_sink.ready.eq(1)


        with m.FSM(domain="ss"):

            # WAIT_FOR_DATA -- we're idly waiting for our input data stream to become valid.
            with m.State("WAIT_FOR_DATA"):

                # Constantly latch in our data parameters until we get a new data packet.
                m.d.ss += [
                    sequence_number  .eq(self.sequence_number),
                    endpoint_number  .eq(self.endpoint_number),
                    data_length      .eq(self.data_length),
                    direction        .eq(self.direction)
                ]

                # Once our data goes valid, begin sending our data.
                with m.If(data_sink.valid.any()):
                    m.next = "SEND_HEADER"

                with m.Elif(self.send_zlp):
                    m.next = "SEND_ZLP"


            # SEND_HEADER -- we're sending the header associated with our data packet.
            with m.State("SEND_HEADER"):
                header = DataHeaderPacket()
                m.d.comb += [
                    header_source.header    .eq(header),
                    header_source.valid     .eq(1),

                    # We're sending a data packet from up to the host.
                    header.type             .eq(HeaderPacketType.DATA),
                    header.direction        .eq(direction),
                    header.device_address   .eq(self.address),

                    # Fill in our input parameters...
                    header.data_sequence    .eq(sequence_number),
                    header.data_length      .eq(data_length),
                    header.endpoint_number  .eq(endpoint_number),
                ]

                # Once our header is accepted, move on to passing through our payload.
                with m.If(header_source.ready):
                    m.next = "SEND_PAYLOAD"


            # SEND_PAYLOAD -- we're now passing our payload data to our transmitter; which will
            # drive ready when it's time to accept data.
            with m.State("SEND_PAYLOAD"):

                # Once our packet is complete, we'll go back to idle.
                with m.If(~data_sink.valid.any()):
                    m.next = "WAIT_FOR_DATA"


            # SEND_ZLP -- we're sending a ZLP; which in our case means we'll be sending a header
            # without driving our data stream.
            with m.State("SEND_ZLP"):
                header = DataHeaderPacket()
                m.d.comb += [
                    header_source.header    .eq(header),
                    header_source.valid     .eq(1),

                    # We're sending a data packet from up to the host.
                    header.type             .eq(HeaderPacketType.DATA),
                    header.direction        .eq(direction),
                    header.device_address   .eq(self.address),

                    # Fill in our input parameters...
                    header.data_sequence    .eq(sequence_number),
                    header.data_length      .eq(0),
                    header.endpoint_number  .eq(endpoint_number),
                ]

                # Once our header is accepted, we can move directly back to idle.
                # Our transmitter will handle generating the zero-length DPP.
                with m.If(header_source.ready):
                    m.next = "WAIT_FOR_DATA"



        return m
